import polars as pl
import pytest

from hstrat._auxiliary_lib import get_sole_scalar_value_polars


def test_get_sole_scalar_value_polars_single_value():
    df = pl.DataFrame({"col": [42, 42, 42]})
    assert 42 == get_sole_scalar_value_polars(df, "col")
    assert 42 == get_sole_scalar_value_polars(df.lazy(), "col")


def test_get_sole_scalar_value_polars_multiple_values_error():
    df = pl.DataFrame({"col": [1, 2, 2]})
    with pytest.raises(ValueError):
        get_sole_scalar_value_polars(df, "col")
    with pytest.raises(ValueError):
        get_sole_scalar_value_polars(df.lazy(), "col")


def test_get_sole_scalar_value_polars_empty_dataframe_error():
    df = pl.DataFrame({"col": []})
    with pytest.raises(ValueError):
        get_sole_scalar_value_polars(df, "col")
    with pytest.raises(ValueError):
        get_sole_scalar_value_polars(df.lazy(), "col")



import pytest

from hstrat._auxiliary_lib import iter_chunks


@pytest.fixture()
def seq():
    return [1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_chunk_size_one(seq):
    for chunk in iter_chunks(seq, 1):
        assert chunk in [[1], [2], [3], [4], [5], [6], [7], [8], [9]]


def test_chunk_size_two(seq):
    for chunk in iter_chunks(seq, 2):
        assert chunk in [[1, 2], [3, 4], [5, 6], [7, 8], [9]]


def test_chunk_size_greater_than_length(seq):
    for chunk in iter_chunks(seq, 10):
        assert chunk == [1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_empty_sequence():
    assert list(iter_chunks([], 5)) == []


def test_non_list_sequence():
    assert list(iter_chunks("123456", 2)) == ["12", "34", "56"]


def test_chunking_integers():
    # check if chunking a list of integers returns the correct chunks
    input_list = [1, 2, 3, 4, 5, 6]
    chunk_size = 2
    expected_output = [[1, 2], [3, 4], [5, 6]]
    output = list(iter_chunks(input_list, chunk_size))
    assert output == expected_output


def test_chunking_strings():
    # check if chunking a list of strings returns the correct chunks
    input_list = ["apple", "banana", "cherry", "date"]
    chunk_size = 2
    expected_output = [["apple", "banana"], ["cherry", "date"]]
    output = list(iter_chunks(input_list, chunk_size))
    assert output == expected_output


def test_chunk_size_larger_than_input_list():
    # check if chunking with chunk size larger than the input list returns the
    # input list
    input_list = [1, 2, 3, 4]
    chunk_size = 10
    expected_output = [[1, 2, 3, 4]]
    output = list(iter_chunks(input_list, chunk_size))
    assert output == expected_output


def test_chunk_size_larger_than_input_list_with_start():
    # check if chunking with chunk size larger than the input list returns the
    # input list
    input_list = [1, 2, 3, 4]
    chunk_size = 10
    expected_output = [[2, 3, 4]]
    output = list(iter_chunks(input_list, chunk_size, 1))
    assert output == expected_output


def test_chunk_size_equal_to_one():
    # check if chunking with chunk size equal to 1 returns a list of singleton
    # lists
    input_list = [1, 2, 3, 4]
    chunk_size = 1
    expected_output = [[1], [2], [3], [4]]
    output = list(iter_chunks(input_list, chunk_size))
    assert output == expected_output


def test_start_parameter():
    # check if the `start` parameter works as expected
    input_list = [1, 2, 3, 4, 5, 6]
    chunk_size = 2
    start = 2
    expected_output = [[3, 4], [5, 6]]
    output = list(iter_chunks(input_list, chunk_size, start))
    assert output == expected_output


def test_chunk_size_does_not_divide_input_length():
    # check if chunking with chunk size that does not divide the input list
    # evenly works as expected
    input_list = [1, 2, 3, 4, 5]
    chunk_size = 2
    expected_output = [
        [1, 2],
        [3, 4],
        [
            5,
        ],
    ]
    output = list(iter_chunks(input_list, chunk_size))
    assert output == expected_output


def test_chunk_size_equal_to_input_length():
    # check if chunking with chunk size equal to the input list length returns
    # the input list as a single chunk
    input_list = [1, 2, 3, 4, 5]
    chunk_size = 5
    expected_output = [[1, 2, 3, 4, 5]]
    output = list(iter_chunks(input_list, chunk_size))
    assert output == expected_output


def test_chunk_size_does_not_divide_input_length_with_start():
    # check if chunking with chunk size that does not divide the input list evenly and a non-zero `start` works as expected
    input_list = [1, 2, 3, 4, 5]
    chunk_size = 2
    start = 1
    expected_output = [[2, 3], [4, 5]]
    output = list(iter_chunks(input_list, chunk_size, start))
    assert output == expected_output


def test_chunk_size_equal_to_input_length_with_start():
    # check if chunking with chunk size equal to the input list length and a
    # non-zero `start` works as expected
    input_list = [1, 2, 3, 4, 5]
    chunk_size = 5
    start = 1
    expected_output = [[2, 3, 4, 5]]
    output = list(iter_chunks(input_list, chunk_size, start))
    assert output == expected_output


def test_chunk_size_divides_input_length_with_start_2():
    # check if chunking with chunk size that divides the input list evenly and
    # a `start` of 2 works as expected
    input_list = [1, 2, 3, 4, 5]
    chunk_size = 2
    start = 2
    expected_output = [
        [3, 4],
        [
            5,
        ],
    ]
    output = list(iter_chunks(input_list, chunk_size, start))
    assert output == expected_output


def test_start_equals_length():
    # check if `start` equal to the length of the input list returns an empty
    # list
    input_list = [1, 2, 3, 4, 5]
    chunk_size = 2
    start = len(input_list)
    expected_output = []
    output = list(iter_chunks(input_list, chunk_size, start))
    assert output == expected_output



import numpy as np
import opytional as opyt
import pandas as pd
import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "estimator",
    ["maximum_likelihood", "unbiased"],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform"],
)
@pytest.mark.parametrize(
    "force_common_ancestry",
    [True, False, None],
)
def test_build_distance_matrix_biopython_empty(
    differentia_bit_width,
    estimator,
    prior,
    force_common_ancestry,
):

    population = []
    dm = hstrat.build_distance_matrix_biopython(
        population,
        estimator,
        prior,
        force_common_ancestry=force_common_ancestry,
    )
    assert dm.matrix == []
    assert dm.names == []


@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "estimator",
    ["maximum_likelihood", "unbiased"],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform"],
)
@pytest.mark.parametrize(
    "force_common_ancestry",
    [True, False, None],
)
@pytest.mark.parametrize(
    "names",
    [["foo"], None],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_build_distance_matrix_biopython_singleton(
    differentia_bit_width, estimator, prior, force_common_ancestry, names, wrap
):

    population = [wrap(hstrat.HereditaryStratigraphicColumn())]
    dm = hstrat.build_distance_matrix_biopython(
        population,
        estimator,
        prior,
        taxon_labels=names,
        force_common_ancestry=force_common_ancestry,
    )
    m = hstrat.build_distance_matrix_numpy(
        population,
        estimator,
        prior,
        force_common_ancestry=force_common_ancestry,
    )
    assert np.array_equal(m, dm.matrix)
    assert opyt.or_value(names, ["0"]) == dm.names


@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "estimator",
    ["maximum_likelihood", "unbiased"],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform"],
)
@pytest.mark.parametrize(
    "names",
    [["foo", "bar"], None],
)
@pytest.mark.parametrize(
    "force_common_ancestry",
    [True, False],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_build_distance_matrix_biopython_pair_disjoint(
    differentia_bit_width,
    estimator,
    prior,
    names,
    force_common_ancestry,
    wrap,
):

    population = [
        wrap(hstrat.HereditaryStratigraphicColumn()),
        wrap(hstrat.HereditaryStratigraphicColumn()),
    ]
    dm = hstrat.build_distance_matrix_biopython(
        population,
        estimator,
        prior,
        taxon_labels=names,
        force_common_ancestry=force_common_ancestry,
    )
    m = hstrat.build_distance_matrix_numpy(
        population,
        estimator,
        prior,
        force_common_ancestry=force_common_ancestry,
    )

    df_dm = pd.DataFrame(dm.matrix, dm.names)

    m[np.triu_indices(m.shape[0], 1)] = np.nan
    df_m = pd.DataFrame(m, opyt.or_value(names, ["0", "1"]))

    assert df_m.equals(df_dm)


@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "estimator",
    ["maximum_likelihood", "unbiased"],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform"],
)
@pytest.mark.parametrize(
    "names",
    [["foo", "bar"], None],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_build_distance_matrix_biopython_pair_disjoint2(
    differentia_bit_width,
    estimator,
    prior,
    names,
    wrap,
):

    population = [
        wrap(hstrat.HereditaryStratigraphicColumn()),
        wrap(hstrat.HereditaryStratigraphicColumn()),
    ]
    with pytest.raises(ValueError):
        hstrat.build_distance_matrix_biopython(
            population,
            estimator,
            prior,
            taxon_labels=names,
            force_common_ancestry=None,
        )


@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
@pytest.mark.parametrize(
    "estimator",
    ["maximum_likelihood", "unbiased"],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform"],
)
@pytest.mark.parametrize(
    "names",
    [["foo", "bar"], None],
)
@pytest.mark.parametrize(
    "force_common_ancestry",
    [True, False],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_build_distance_matrix_biopython_pair_commonancestry(
    differentia_bit_width,
    policy,
    estimator,
    prior,
    names,
    force_common_ancestry,
    wrap,
):

    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_bit_width,
    ).CloneNthDescendant(7)
    population = [
        wrap(common_ancestor.CloneNthDescendant(4)),
        wrap(common_ancestor.CloneNthDescendant(9)),
    ]
    dm = hstrat.build_distance_matrix_biopython(
        population,
        estimator,
        prior,
        taxon_labels=names,
        force_common_ancestry=force_common_ancestry,
    )
    m = hstrat.build_distance_matrix_numpy(
        population,
        estimator,
        prior,
        force_common_ancestry=force_common_ancestry,
    )

    df_dm = pd.DataFrame(dm.matrix, dm.names)

    m[np.triu_indices(m.shape[0], 1)] = np.nan
    df_m = pd.DataFrame(m, opyt.or_value(names, ["0", "1"]))

    assert df_m.equals(df_dm)


@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
@pytest.mark.parametrize(
    "estimator",
    ["maximum_likelihood", "unbiased"],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform"],
)
@pytest.mark.parametrize(
    "names",
    [["foo", "bar"]],
)
@pytest.mark.parametrize(
    "force_common_ancestry",
    [True, False],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_build_distance_matrix_biopython_pair_iternames(
    differentia_bit_width,
    policy,
    estimator,
    prior,
    names,
    force_common_ancestry,
    wrap,
):

    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_bit_width,
    ).CloneNthDescendant(7)
    population = [
        wrap(common_ancestor.CloneNthDescendant(4)),
        wrap(common_ancestor.CloneNthDescendant(9)),
    ]
    dm = hstrat.build_distance_matrix_biopython(
        population,
        estimator,
        prior,
        taxon_labels=opyt.apply_if(names, iter),
        force_common_ancestry=force_common_ancestry,
    )
    m = hstrat.build_distance_matrix_numpy(
        population,
        estimator,
        prior,
        force_common_ancestry=force_common_ancestry,
    )

    df_dm = pd.DataFrame(dm.matrix, dm.names)

    m[np.triu_indices(m.shape[0], 1)] = np.nan
    df_m = pd.DataFrame(m, opyt.or_value(names, ["0", "1"]))

    assert df_m.equals(df_dm)


@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(4),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
@pytest.mark.parametrize(
    "estimator",
    ["maximum_likelihood", "unbiased"],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform"],
)
def test_build_distance_matrix_biopython_pair_artifact_types_equiv(
    policy, estimator, prior
):

    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
    ).CloneNthDescendant(7)
    population = [
        common_ancestor.CloneNthDescendant(4),
        common_ancestor.CloneNthDescendant(9),
    ]
    dm_column = hstrat.build_distance_matrix_biopython(
        population,
        estimator,
        prior,
    )
    dm_specimen = hstrat.build_distance_matrix_biopython(
        [*map(hstrat.col_to_specimen, population)],
        estimator,
        prior,
    )

    df_dm_column = pd.DataFrame(dm_column.matrix, dm_column.names)
    df_dm_specimen = pd.DataFrame(dm_specimen.matrix, dm_specimen.names)

    assert df_dm_column.equals(df_dm_specimen)



import typing

caretdown_marker: typing.List[typing.Tuple[int, int]] = [
    (0, 0),
    (-1, 3**0.5),
    (1, 3**0.5),
    (0, 0),
]
"""A equilateral triangle, oriented tip down and positioned within the
encapsulating unit cell so that it appears just above the coordinates it is
placed at.

Equivalent to matplotlib "CARETDOWN" marker, except that it is a closed path so
edge outline will appear properly.
"""



import pytest

from hstrat.phylogenetic_inference.tree._impl import (
    Searchtable,
    SearchtableRecord,
)


@pytest.fixture
def empty_searchtable() -> Searchtable:
    return Searchtable()


@pytest.fixture
def populated_searchtable() -> Searchtable:
    st = Searchtable()
    # Create a simple tree:
    #    0 (root)
    #   / \
    #  1   2 (inner)
    #     / \
    #    3   4
    st.create_offspring(parent_id=0, differentia=None, rank=1, taxon_label="A")
    st.create_offspring(parent_id=0, differentia=102, rank=1, taxon_label="in")
    st.create_offspring(parent_id=2, differentia=None, rank=2, taxon_label="C")
    st.create_offspring(parent_id=2, differentia=None, rank=2, taxon_label="D")
    return st


def test_init(empty_searchtable: Searchtable):
    records = empty_searchtable.to_records()
    assert len(records) == 1
    assert records[0]["taxon_id"] == 0
    assert records[0]["taxon_label"] == "_root"
    assert records[0]["rank"] == 0
    # The root node has itself as search_ancestor_id and search_next_sibling_id
    assert records[0]["search_ancestor_id"] == 0
    assert records[0]["search_next_sibling_id"] == 0
    assert records[0]["search_first_child_id"] == 0


def test_get_differentia_of(populated_searchtable: Searchtable):
    assert populated_searchtable.get_differentia_of(1) is None
    assert populated_searchtable.get_differentia_of(2) == 102


def test_get_rank_of(populated_searchtable: Searchtable):
    assert populated_searchtable.get_rank_of(1) == 1
    assert populated_searchtable.get_rank_of(3) == 2


def test_to_records(populated_searchtable: Searchtable):
    records = populated_searchtable.to_records()
    assert isinstance(records, list)
    assert all(isinstance(r, dict) for r in records)
    node_2 = records[2]
    assert node_2["taxon_id"] == 2
    assert node_2["taxon_label"] == "in"
    assert node_2["differentia"] == 102
    assert node_2["rank"] == 1


def test_has_search_parent(populated_searchtable: Searchtable):
    # The root (0) should have no search parent
    assert not populated_searchtable.has_search_parent(0)
    # All other nodes should have search parents
    for nid in [1, 2, 3, 4]:
        assert populated_searchtable.has_search_parent(nid)


def test_iter_search_children_of(populated_searchtable: Searchtable):
    """Test iterating over search children of a given node."""
    children_of_0 = {*populated_searchtable.iter_search_children_of(0)}
    assert set(children_of_0) == {1, 2}

    children_of_2 = {*populated_searchtable.iter_search_children_of(2)}
    assert set(children_of_2) == {3, 4}

    # Node 3 is a leaf, so no children
    children_of_3 = {*populated_searchtable.iter_search_children_of(3)}
    assert children_of_3 == set()


def test_iter_inner_search_children_of(populated_searchtable: Searchtable):
    """Test iter_inner_search_children_of (excluding leaf nodes)."""
    inner_children_of_0 = {
        *populated_searchtable.iter_inner_search_children_of(0),
    }
    assert inner_children_of_0 == {2}

    inner_children_of_2 = {
        *populated_searchtable.iter_inner_search_children_of(2),
    }
    assert inner_children_of_2 == set()


def test_attach_search_parent(empty_searchtable: Searchtable):
    # Create a node without attaching parent first (directly manipulate records)
    new_taxon_id = len(empty_searchtable._records)
    new_node = SearchtableRecord(
        rank=1,
        taxon_id=new_taxon_id,
        taxon_label="X",
        ancestor_id=0,
        differentia=999,
        search_first_child_id=new_taxon_id,
        search_next_sibling_id=new_taxon_id,
        search_ancestor_id=new_taxon_id,
    )
    empty_searchtable._records.append(new_node)

    # Before attaching
    assert not empty_searchtable.has_search_parent(new_taxon_id)

    # Attach the parent (root: 0)
    empty_searchtable.attach_search_parent(taxon_id=new_taxon_id, parent_id=0)

    # After attaching, should have a parent
    assert empty_searchtable.has_search_parent(new_taxon_id)

    # Check the state with get_records
    records = empty_searchtable.to_records()
    attached_node = records[new_taxon_id]
    assert attached_node["search_ancestor_id"] == 0
    # The parent's first_child should now be new_taxon_id
    assert records[0]["search_first_child_id"] == new_taxon_id
    # The node's next_sibling_id should point to itself if it is the only child
    assert attached_node["search_next_sibling_id"] == new_taxon_id


def test_detach_search_parent(populated_searchtable: Searchtable):
    # Detach node 1 from its parent (0)
    populated_searchtable.detach_search_parent(1)

    # After detachment, node 1 should not have a search parent
    assert not populated_searchtable.has_search_parent(1)

    # Also check that the parent's first_child_id is updated
    records = populated_searchtable.to_records()

    # Previously, 0 had first_child_id as 1 or 2 (depending on insertion order).
    # After detaching 1, the parent's first_child_id should not be 1.
    parent_rec = records[0]
    assert parent_rec["search_first_child_id"] != 1

    # Detached node 1 should have itself as its own ancestor and next_sibling
    detached = records[1]
    assert detached["search_ancestor_id"] == 1
    assert detached["search_next_sibling_id"] == 1


def test_create_offspring(empty_searchtable: Searchtable):
    # Create a single child of root
    cid1 = empty_searchtable.create_offspring(
        parent_id=0, differentia=10, rank=1, taxon_label="Child1"
    )
    records = empty_searchtable.to_records()
    child = records[cid1]
    assert child["ancestor_id"] == 0
    assert child["differentia"] == 10
    assert child["rank"] == 1
    assert child["taxon_label"] == "Child1"
    assert child["search_ancestor_id"] == 0  # after attach_search_parent call
    assert child["search_first_child_id"] == cid1
    assert child["search_next_sibling_id"] == cid1

    # Root should now have a first_child_id equal to the child's ID
    root = records[0]
    assert root["search_first_child_id"] == cid1

    # Create a sibling for the first child
    cid2 = empty_searchtable.create_offspring(
        parent_id=0, differentia=20, rank=1, taxon_label="Child2"
    )
    records = empty_searchtable.to_records()
    root = records[0]
    child1 = records[cid1]
    child2 = records[cid2]

    assert root["search_first_child_id"] == cid2
    assert child2["search_next_sibling_id"] == cid1
    assert child1["search_next_sibling_id"] == cid1



import typing

# use string alias due to circular module initialization issue
HereditaryStratigraphicArtifact = typing.Union[
    "HereditaryStratigraphicColumn",  # noqa: F821
    "HereditaryStratigraphicSpecimen",  # noqa: F821
]



import itertools as it
import typing

import matplotlib as mpl
from matplotlib import pyplot as plt


def mrca_uncertainty_absolute_barplot(
    stratum_retention_policy: typing.Any,
    num_generations: int,
    do_show: bool = False,
    ax: typing.Optional[plt.matplotlib.axes.Axes] = None,
) -> plt.matplotlib.axes.Axes:
    """Profile distribution of retained rank spacing under a retention policy.

    Plots absolute uncertainty for MRCA estimation over column ranks
    (positions) in a hereditary stratigraphic column at a particular generation
    under a particular stratum retention policy.

    Parameters
    ----------
    stratum_retention_policy : Callable
        Object specifying stratum retention policy.
    num_generations : int
        Number of generations to plot.
    ax : matplotlib/pylab axes, optional
        If a valid matplotlib.axes.Axes instance, the plot is drawn in that
        Axes. By default (None), a new axes is created.
    do_show : bool, optional
        Whether to show() the plot automatically.
    """
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
    elif not isinstance(ax, plt.matplotlib.axes.Axes):
        raise ValueError(f"Invalid argument for ax: {ax}")

    xs = [0]
    ys = [0]
    for gen in range(1, num_generations):
        xs.append(gen)
        if stratum_retention_policy.CalcMrcaUncertaintyAbsExact is not None:
            ys.append(
                stratum_retention_policy.CalcMrcaUncertaintyAbsExact(
                    num_generations,
                    num_generations,
                    gen,
                )
            )
        else:
            # not implemented for policies without CalcMrcaUncertaintyAbsExact
            ys.append(float("nan"))

    ax.bar(
        xs,
        ys,
        color=[
            *it.islice(
                it.cycle(
                    [
                        "mediumpurple",
                        "mediumslateblue",
                    ]
                ),
                num_generations,
            )
        ],
        width=1.0,
    )
    ax.set_xlabel("Position (Rank)")
    ax.set_ylabel("Absolute MRCA\n Uncertainty", labelpad=20)
    ax.xaxis.set_major_locator(
        mpl.ticker.MaxNLocator(
            nbins="auto",
            steps=[1, 2, 5, 10],
            integer=True,
            min_n_ticks=0,
        )
    )
    ax.yaxis.set_major_locator(
        mpl.ticker.MaxNLocator(
            nbins="auto",
            steps=[1, 2, 5, 10],
            integer=True,
            min_n_ticks=0,
        )
    )

    # strip any negative xticks
    ax.set_xticks([tick for tick in ax.get_xticks() if tick >= 0])

    if do_show:
        plt.show()

    return ax



"""Implementation helpers."""

from ._calc_rank_of_first_retained_disparity_between import (
    calc_rank_of_first_retained_disparity_between,
)
from ._calc_rank_of_first_retained_disparity_between_bsearch import (
    calc_rank_of_first_retained_disparity_between_bsearch,
)
from ._calc_rank_of_first_retained_disparity_between_generic import (
    calc_rank_of_first_retained_disparity_between_generic,
)
from ._calc_rank_of_last_retained_commonality_between import (
    calc_rank_of_last_retained_commonality_between,
)
from ._calc_rank_of_last_retained_commonality_between_bsearch import (
    calc_rank_of_last_retained_commonality_between_bsearch,
)
from ._calc_rank_of_last_retained_commonality_between_generic import (
    calc_rank_of_last_retained_commonality_between_generic,
)
from ._calc_rank_of_parity_segue_between import (
    calc_rank_of_parity_segue_between,
)
from ._iter_mutual_ranks import iter_mutual_ranks
from ._iter_ranks_of_retained_commonality_between import (
    iter_ranks_of_retained_commonality_between,
)

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "calc_rank_of_first_retained_disparity_between",
    "calc_rank_of_first_retained_disparity_between_bsearch",
    "calc_rank_of_first_retained_disparity_between_generic",
    "calc_rank_of_last_retained_commonality_between",
    "calc_rank_of_last_retained_commonality_between_bsearch",
    "calc_rank_of_last_retained_commonality_between_generic",
    "calc_rank_of_parity_segue_between",
    "iter_mutual_ranks",
    "iter_ranks_of_retained_commonality_between",
]



"""Convenient library repack as a flat namespace."""

from .. import dataframe as _dataframe
from .. import frozen_instrumentation as _frozen_instrumentation
from .. import genome_instrumentation as _genome_instrumentation
from .. import juxtaposition as _juxtaposition
from .. import phylogenetic_inference as _phylogenetic_inference
from .. import serialization as _serialization
from .. import stratum_retention_strategy as _stratum_retention_strategy
from .. import stratum_retention_viz as _stratum_retention_viz
from .. import test_drive as _test_drive
from .._auxiliary_lib import get_hstrat_version as _get_hstrat_version

_top_level_modules = (
    _dataframe,
    _frozen_instrumentation,
    _genome_instrumentation,
    _juxtaposition,
    _phylogenetic_inference,
    _serialization,
    _stratum_retention_strategy,
    _stratum_retention_viz,
    _test_drive,
)

__version__ = _get_hstrat_version()

__all__ = tuple(
    _item for _module in _top_level_modules for _item in _module.__all__
)


def __dir__() -> list:
    """Returns the entire symbol list."""
    return __all__


def __getattr__(name: str) -> object:
    """Allows for equivalent behavior to having `from mod import *`
    in this file without the load-time cost of star imports.
    Previously, this file eagerly imported every symbol from the
    below packages. Now, the symbol is only imported if determined
    to be in said package.
    """
    for _module in _top_level_modules:
        if name in _module.__all__:
            return getattr(_module, name)
    raise AttributeError(f"Flat namespace 'hstrat' has no attribute '{name}'")



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec
from .._impl import calc_provided_uncertainty


class IterRetainedRanks:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "IterRetainedRanks",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "IterRetainedRanks", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "IterRetainedRanks",
        policy: PolicyCouplerBase,
        num_strata_deposited: int,
    ) -> typing.Iterator[int]:
        """Iterate over retained strata ranks at `num_strata_deposited` in
        ascending order."""
        spec = policy.GetSpec()
        guaranteed_resolution = spec.GetDepthProportionalResolution()

        if num_strata_deposited < guaranteed_resolution * 2 + 1:
            # use identity mapping before first ranks are condemned
            yield from range(num_strata_deposited)
            return

        cur_stage_uncertainty = calc_provided_uncertainty(
            guaranteed_resolution,
            num_strata_deposited,
        )
        cur_stage_max_idx = (  # noqa: F841, keep unused for comprehensibility
            num_strata_deposited // cur_stage_uncertainty
        )

        prev_stage_uncertainty = cur_stage_uncertainty // 2
        prev_stage_max_idx = (
            num_strata_deposited - 2
        ) // prev_stage_uncertainty
        thresh_idx = (
            2 * prev_stage_max_idx - 4 * guaranteed_resolution + 2
        ) // 2

        # note that cur stage uncertainty is iterated through first
        # because ranks are removed from the back, the old prev stage
        # uncertainty lingers at more recent ranks
        yield from range(
            0,
            thresh_idx * cur_stage_uncertainty,
            cur_stage_uncertainty,
        )
        yield from range(
            thresh_idx * cur_stage_uncertainty,
            num_strata_deposited,
            prev_stage_uncertainty,
        )

        last_rank = num_strata_deposited - 1
        if thresh_idx * cur_stage_uncertainty > last_rank:
            if last_rank > 0 and last_rank % cur_stage_uncertainty:
                yield last_rank
        elif last_rank > 0 and last_rank % prev_stage_uncertainty:
            yield last_rank



"""Configuration file for pytest."""

import random
import warnings

import numpy as np
import pkg_resources
import pytest


def pytest_configure(config):
    # check for testing requirements
    # provide install instructions if they're missing
    try:
        from hstrat._auxiliary_lib._check_testing_requirements import (
            check_testing_requirements,
        )

        check_testing_requirements()
    except (
        ModuleNotFoundError,
        pkg_resources.DistributionNotFound,
        pkg_resources.VersionConflict,
    ) as exception:
        warnings.warn(
            "Missing some testing requirements. "
            "For instructions on installing dev dependencies, see "
            "https://hstrat.readthedocs.io/en/latest/contributing.html "
            "or CONTRIBUTING.rst"
            "\n"
            '  (Warning caused by exception "' + str(exception) + '".)'
        )


@pytest.fixture(autouse=True)
def reseed_random(request: pytest.FixtureRequest) -> None:
    """Reseed random number generator to ensure determinstic test."""
    random.seed(request.node.name)
    np.random.seed(random.randint(1, 2**32))



import numpy as np
import pandas as pd

from .._auxiliary_lib import numpy_fromiter_polyfill
from ..frozen_instrumentation import HereditaryStratigraphicSpecimen
from ..genome_instrumentation import HereditaryStratigraphicColumn


def col_to_specimen(
    column: HereditaryStratigraphicColumn,
) -> HereditaryStratigraphicSpecimen:
    """Create a postprocessing representation of the differentia retained
    by an extant HereditaryStratigraphicColumn, indexed by deposition rank."""
    differentia = numpy_fromiter_polyfill(
        column.IterRetainedDifferentia(),
        dtype=np.min_scalar_type(
            2 ** column.GetStratumDifferentiaBitWidth() - 1
        ),
    )
    ranks = numpy_fromiter_polyfill(
        column.IterRetainedRanks(),
        dtype=np.min_scalar_type(column.GetNumStrataDeposited() - 1),
    )

    return HereditaryStratigraphicSpecimen(
        stratum_differentia_series=pd.Series(data=differentia, index=ranks),
        stratum_differentia_bit_width=column.GetStratumDifferentiaBitWidth(),
    )



import itertools as it
import random

import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreDict,
            marks=pytest.mark.heavy,
        ),
        hstrat.HereditaryStratumOrderedStoreList,
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_comparison_commutativity_asynchronous(
    retention_policy,
    ordered_store,
    wrap,
):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert (
                hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(first), wrap(second)], prior="arbitrary"
                )
                == hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(second), wrap(first)], prior="arbitrary"
                )
                == hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(second), wrap(first), wrap(first)], prior="arbitrary"
                )
                == hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(second), wrap(first)] * 3, prior="arbitrary"
                )
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].CloneDescendant()
        for individual in population:
            # asynchronous generations
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        pytest.param(
            hstrat.perfect_resolution_algo.Policy(),
            marks=pytest.mark.heavy_2a,
        ),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreDict,
            marks=pytest.mark.heavy_2b,
        ),
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_comparison_commutativity_synchronous(
    retention_policy,
    ordered_store,
    wrap,
):

    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert (
                hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(first), wrap(second)], prior="arbitrary"
                )
                == hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(second), wrap(first)], prior="arbitrary"
                )
                == hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(second), wrap(first), wrap(second)],
                    prior="arbitrary",
                )
                == hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(second), wrap(first)] * 3, prior="arbitrary"
                )
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        # synchronous generations
        for individual in population:
            individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreDict,
            marks=pytest.mark.heavy,
        ),
        hstrat.HereditaryStratumOrderedStoreList,
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_comparison_validity(retention_policy, ordered_store, wrap):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            assert (
                hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(first), wrap(second)], prior="arbitrary"
                )
                >= 0
            )
        for first, second, third in it.islice(
            it.combinations(population, 3), 100
        ):
            assert (
                hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(first), wrap(second), wrap(third)],
                    prior="arbitrary",
                )
                >= 0
            )

        # advance generations asynchronously
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        for individual in population:
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy1",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "retention_policy2",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_scenario_no_mrca(
    retention_policy1,
    retention_policy2,
    ordered_store,
    wrap,
):
    first = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy1,
    )
    second = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy2,
    )

    for _generation in range(100):
        _ = _generation
        assert (
            hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(first), wrap(second)],
                prior="arbitrary",
            )
            == 0
        )
        assert (
            hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(second), wrap(first)],
                prior="arbitrary",
            )
            == 0
        )
        assert (
            hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(first), wrap(second), wrap(first)],
                prior="arbitrary",
            )
            == 0
        )
        assert (
            hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(second), wrap(first)] * 3,
                prior="arbitrary",
            )
            == 0
        )

        first.DepositStratum()
        second.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreDict,
            marks=pytest.mark.heavy,
        ),
        hstrat.HereditaryStratumOrderedStoreList,
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_scenario_no_divergence(retention_policy, ordered_store, wrap):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )

    for _generation in range(100):
        _ = _generation
        assert (
            hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(column), wrap(column)],
                prior="arbitrary",
            )
            == 0
        )
        assert (
            hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(column), wrap(column), wrap(column)],
                prior="arbitrary",
            )
            == 0
        )
        assert (
            hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(column), wrap(column)] * 3, prior="arbitrary"
            )
            == 0
        )
        column.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.8, 0.95],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        pytest.param(
            hstrat.col_to_specimen,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_CalcRankOfMrcaUncertaintyWith_narrow_shallow(
    retention_policy,
    differentia_width,
    confidence_level,
    wrap,
):

    columns = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
            stratum_ordered_store=hstrat.HereditaryStratumOrderedStoreDict,
        )
        for __ in range(20)
    ]

    steps = list(
        range(
            columns[
                0
            ].CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=1 - confidence_level,
            )
            - columns[0].GetNumStrataDeposited()
        )
    )

    for step1, step2 in it.product(steps, steps):
        column1 = [col.Clone() for col in columns]
        column2 = [col.Clone() for col in columns]
        column3 = [col.Clone() for col in columns]
        for __ in range(step1):
            for col in column1:
                col.DepositStratum()
        for __ in range(step2):
            for col in column2:
                col.DepositStratum()
        for __ in range(step1):
            for col in column2:
                col.DepositStratum()

        for c1, c2 in zip(column1, column2):
            assert (
                hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(c1), wrap(c2)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                is None
            )
            assert (
                hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(c2), wrap(c1)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                is None
            )
        for c1, c2, c3 in zip(column1, column2, column3):
            assert (
                hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(c1), wrap(c2), wrap(c3)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                is None
            )
            assert (
                hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(c2), wrap(c3), wrap(c1)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                is None
            )


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.8, 0.95],
)
@pytest.mark.parametrize(
    "mrca_rank",
    [100],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        pytest.param(
            hstrat.col_to_specimen,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_CalcRankOfMrcaUncertaintyWith_narrow_with_mrca(
    retention_policy,
    differentia_width,
    confidence_level,
    mrca_rank,
    wrap,
):

    columns = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
            stratum_ordered_store=hstrat.HereditaryStratumOrderedStoreDict,
        )
        for __ in range(20)
    ]

    for column in columns:
        column.DepositStrata(mrca_rank)

    steps = (0, 16, 51)

    for step1, step2 in it.product(steps, steps):
        column1 = [col.Clone() for col in columns]
        column2 = [col.Clone() for col in columns]
        for __ in range(step1):
            for col in column1:
                col.DepositStratum()
        for i in range(step2):
            for col in column2:
                col.DepositStratum()

        for c1, c2 in zip(column1, column2):
            assert hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(c1), wrap(c2)],
                prior="arbitrary",
                confidence_level=confidence_level,
            ) == hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(c2), wrap(c1)],
                prior="arbitrary",
                confidence_level=confidence_level,
            )
            res = hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(c1), wrap(c2)],
                prior="arbitrary",
                confidence_level=confidence_level,
            )

            if res is not None:
                assert res >= 0
                assert hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(c1), wrap(c2)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                ) >= hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(c2), wrap(c1)],
                    prior="arbitrary",
                    confidence_level=confidence_level / 2,
                )

    for step1, step2 in it.product(steps, steps):
        column1 = [col.Clone() for col in columns]
        column2 = [col.Clone() for col in columns]
        column3 = [col.Clone() for col in columns]
        for __ in range(step1):
            for col in column1:
                col.DepositStratum()
        for i in range(step2):
            for col in column2:
                col.DepositStratum()
        for i in range(step1):
            for col in column3:
                col.DepositStratum()

        for c1, c2, c3 in zip(column1, column2, column3):
            assert hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(c1), wrap(c2), wrap(c3)],
                prior="arbitrary",
                confidence_level=confidence_level,
            ) == hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(c2), wrap(c1), wrap(c3)],
                prior="arbitrary",
                confidence_level=confidence_level,
            )
            res = hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(c1), wrap(c2), wrap(c3)],
                prior="arbitrary",
                confidence_level=confidence_level,
            )

            if res is not None:
                assert res >= 0
                assert hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(c1), wrap(c2), wrap(c3)],
                    prior="arbitrary",
                    confidence_level=confidence_level,
                ) >= hstrat.calc_rank_of_mrca_uncertainty_among(
                    [wrap(c2), wrap(c1), wrap(c3)],
                    prior="arbitrary",
                    confidence_level=confidence_level / 2,
                )


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.8, 0.95],
)
@pytest.mark.parametrize(
    "mrca_rank",
    [0, 100],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        pytest.param(
            hstrat.col_to_specimen,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_CalcRankOfMrcaUncertaintyWith_narrow_no_mrca(
    retention_policy,
    differentia_width,
    confidence_level,
    mrca_rank,
    wrap,
):
    def make_column():
        return hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
            stratum_ordered_store=hstrat.HereditaryStratumOrderedStoreDict,
        )

    columns = [make_column() for __ in range(20)]

    for column in columns:
        column.DepositStrata(mrca_rank)

    steps = (0, 16, 51)

    for step1, step2 in it.product(steps, steps):
        column1 = [make_column() for col in columns]
        column2 = [make_column() for col in columns]
        column3 = [make_column() for col in columns]
        for __ in range(step1):
            for col in column1:
                col.DepositStratum()
        for i in range(step2):
            for col in column2:
                col.DepositStratum()
        for i in range(step1):
            for col in column3:
                col.DepositStratum()

        for c1, c2 in zip(column1, column2):
            assert hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(c1), wrap(c2)],
                prior="arbitrary",
                confidence_level=confidence_level,
            ) == hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(c2), wrap(c1)],
                prior="arbitrary",
                confidence_level=confidence_level,
            )
            res = hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(c1), wrap(c2)],
                prior="arbitrary",
                confidence_level=confidence_level,
            )

            if res is not None:
                assert 0 <= res
                assert (
                    hstrat.calc_rank_of_mrca_uncertainty_among(
                        [wrap(c1), wrap(c2)],
                        prior="arbitrary",
                        confidence_level=confidence_level,
                    )
                    >= hstrat.calc_rank_of_mrca_uncertainty_among(
                        [wrap(c2), wrap(c1)],
                        prior="arbitrary",
                        confidence_level=confidence_level / 2,
                    )
                    or hstrat.calc_rank_of_mrca_uncertainty_among(
                        [wrap(c1), wrap(c2)],
                        prior="arbitrary",
                        confidence_level=confidence_level,
                    )
                    == 0
                )

        for c1, c2, c3 in zip(column1, column2, column3):
            assert hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(c1), wrap(c2), wrap(c3)],
                prior="arbitrary",
                confidence_level=confidence_level,
            ) == hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(c3), wrap(c1), wrap(c2)],
                prior="arbitrary",
                confidence_level=confidence_level,
            )
            res = hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(c1), wrap(c2), wrap(c3)],
                prior="arbitrary",
                confidence_level=confidence_level,
            )

            if res is not None:
                assert 0 <= res
                assert (
                    hstrat.calc_rank_of_mrca_uncertainty_among(
                        [wrap(c1), wrap(c2), wrap(c3)],
                        prior="arbitrary",
                        confidence_level=confidence_level,
                    )
                    >= hstrat.calc_rank_of_mrca_uncertainty_among(
                        [wrap(c1), wrap(c2), wrap(c3)],
                        prior="arbitrary",
                        confidence_level=confidence_level / 2,
                    )
                    or hstrat.calc_rank_of_mrca_uncertainty_among(
                        [wrap(c1), wrap(c2), wrap(c3)],
                        prior="arbitrary",
                        confidence_level=confidence_level,
                    )
                    == 0
                )


@pytest.mark.filterwarnings("ignore:Empty or singleton population.")
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_calc_rank_of_mrca_uncertainty_among_singleton(wrap):
    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    for __ in range(10):
        assert (
            hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(c1)], prior="arbitrary"
            )
            is None
        )
        c1.DepositStratum()

    c2 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    for __ in range(10):
        assert (
            hstrat.calc_rank_of_mrca_uncertainty_among(
                [wrap(c1)], prior="arbitrary"
            )
            is None
        )
        c2.DepositStratum()


@pytest.mark.filterwarnings("ignore:Empty or singleton population.")
def test_calc_rank_of_mrca_uncertainty_among_empty():
    assert (
        hstrat.calc_rank_of_mrca_uncertainty_among([], prior="arbitrary")
        is None
    )


@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_calc_rank_of_mrca_uncertainty_among_generator(wrap):
    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    for __ in range(10):
        assert hstrat.calc_rank_of_mrca_uncertainty_among(
            [wrap(c1) for __ in range(10)], prior="arbitrary"
        ) == hstrat.calc_rank_of_mrca_uncertainty_among(
            (wrap(c1) for __ in range(10)), prior="arbitrary"
        )
        c1.DepositStratum()

    c2 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    for __ in range(10):
        assert hstrat.calc_rank_of_mrca_uncertainty_among(
            [wrap(c2) for __ in range(10)], prior="arbitrary"
        ) == hstrat.calc_rank_of_mrca_uncertainty_among(
            (wrap(c2) for __ in range(10)), prior="arbitrary"
        )
        c2.DepositStratum()



import random
import typing


def random_choice_generator(sequence: typing.Sequence) -> typing.Any:
    """A generator that yields an infinite stream of randomly-selected elements
    from a provided sequence.

    Parameters
    ----------
    sequence : iterable
        The sequence from which to randomly select elements.

    Yields
    ------
    object
        Randomly-selected elements from the sequence.

    Examples
    --------
    >>> stream = random_choice_generator([1, 2, 3])
    >>> next(stream)
    1
    >>> next(stream)
    3
    >>> next(stream)
    2
    >>> next(stream)
    1
    >>> next(stream)
    3
    >>> next(stream)
    2
    """
    while True:
        yield random.choice(sequence)



import argparse
import functools
import logging

from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli

from .. import hstrat
from .._auxiliary_lib import (
    configure_prod_logging,
    format_cli_description,
    get_hstrat_version,
    log_context_duration,
)
from ._surface_postprocess_trie import surface_postprocess_trie

raw_message = """Second component of raw interface to tree reconstruction for surface-based genome annotations.
Reads raw phylogeny reconstruction data from reconstruction in tabular data file(s) and writes postprocessed phylogeny data to output file in alife standard format.


End-users are recommended to prefer `hstrat.dataframe.build_tree` for end-to-end processing.
Uses naive method to estimate taxon `origin_time`s and performs no post-processing operations.
Surface-only implementation --- not compatible with HereditaryStratigraphicColumn genome annotations.


Input Schema: Required Columns
==============================
'id' : integer
    - Unique identifier for each taxon (RE alife standard format).
'ancestor_id' : integer
    - Unique identifier for ancestor taxon  (RE alife standard format).
'hstrat_rank' : integer
    - Num generations elapsed for ancestral differentia.
    - Corresponds to `dstream_Tbar` for inner nodes.
    - Corresponds `dstream_T` - 1 for leaf nodes
'hstrat_differentia_bitwidth' : integer
    - Size of annotation differentiae, in bits.
    - Corresponds to `dstream_value_bitwidth`.
'dstream_S' : integer
    - Capacity of dstream buffer used for hstrat surface, in number
        of data items (i.e., differentia values).


If required columns are not included in raw data, they must be created through CLI flags.
For example,
    --with-column 'pl.lit(16).alias("dstream_S")'


Input Schema: Optional Columns
==============================
'dstream_data_id' : integer
    Unique identifier for each taxon from source genome dataframe.


Additional user-provided columns will be forwarded to phylogeny output.
For these columns, output rows for tip nodes are assigned values from corresponding genome row in original data.


Output Schema: Required Columns
===============================
'id' : integer
    Unique identifier for each taxon (RE alife standard format).
    Will be reassigned for contiguity and may differ from input.

'ancestor_id' : integer
    Unique identifier for ancestor taxon (RE alife standard format).

'hstrat_rank_from_t0' : integer
    - Num generations elapsed for ancestral differentia.
    - Corresponds to `dstream_Tbar` - `dstream_S` for inner nodes.
    - Corresponds `dstream_T` - 1 - `dstream_S` for leaf nodes


Output Schema: Optional Columns
===============================
'origin_time' : floating point or integer
    Estimated num generations elapsed from founding ancestor.
    Value depends on the trie postprocessor used.

Additional user-defined columns will be forwarded from the input DataFrame.
Note that the alife-standard `ancestor_list` column is not included in output schema.


See Also
========
-m hstrat.dataframe.surface_postprocess_trie :
    Creates raw phylogeny data postprocessed here.

-m hstrat._auxiliary_lib._alifestd_try_add_ancestor_list_col :
    Adds alife-standard `ancestor_list` column to phylogeny data.


Additional Notes
================
Behind the scenes, the following postprocessing steps occur:
1. Trunk nodes with rank less than `dstream_S` are deleted.

2. Internal unifurcations are collapsed.

3. Taxon `id` values are reassigned.

4. Supplied `trie_postprocessor` functor is applied.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=format_cli_description(raw_message),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    _add_parser_base(
        parser=parser,
        dfcli_module="hstrat.dataframe.surface_postprocess_trie",
        dfcli_version=get_hstrat_version(),
    )
    parser.add_argument(
        "--trie-postprocessor",
        type=str,
        default="hstrat.NopTriePostprocessor()",
        help=(
            "Functor to apply to finalize postprocessed phylogeny. "
            "Will be `eval`-ed as Python code before use."
            "Must support Pandas dataframe input."
        ),
    )
    return parser


if __name__ == "__main__":
    configure_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()

    logging.info(
        f"instantiating trie postprocess functor: "
        f"`{args.trie_postprocessor}`",
    )
    trie_postprocessor = eval(args.trie_postprocessor, {"hstrat": hstrat})

    with log_context_duration(
        "hstrat.dataframe.surface_postprocess_trie", logging.info
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=functools.partial(
                surface_postprocess_trie,
                trie_postprocessor=trie_postprocessor,
            ),
        )



import bitstring


def popcount(val: int) -> int:
    """Count the number of set bits in the binary representation of a
    non-negative int.
    """

    assert val >= 0
    try:
        return val.bit_count()
    except AttributeError:
        return bitstring.Bits(
            # int casts makes robust to numpy int64 etc.
            uint=int(val),
            length=int(val + 1).bit_length(),
        ).count(1)



import typing

import pandas as pd
import polars as pl

from ...._auxiliary_lib import (
    AnyTreeFastPreOrderIter,
    anytree_iterative_deepcopy,
)
from ...priors import ArbitraryPrior
from ...priors._detail import PriorBase
from .._impl import TrieInnerNode, TrieLeafNode
from ._detail import TriePostprocessorBase


def _call_anytree(
    ftor: "AssignOriginTimeNaiveTriePostprocessor",
    trie: TrieInnerNode,
    progress_wrap: typing.Callable,
) -> TrieInnerNode:
    for node in progress_wrap(AnyTreeFastPreOrderIter(trie)):
        if node.is_leaf:
            setattr(node, ftor._assigned_property, node.rank)
            assert isinstance(node, TrieLeafNode)
        elif node.parent is None:
            setattr(node, ftor._assigned_property, 0)
        else:
            interval_mean = ftor._prior.CalcIntervalConditionedMean(
                node.rank,
                min(
                    (
                        child.rank
                        for child in node.children
                        if not child.is_leaf
                    ),
                    default=node.rank + 1,
                ),  # endpoint is exclusive
            )
            setattr(
                node,
                ftor._assigned_property,
                min(
                    interval_mean,
                    min(
                        (child.rank for child in node.children),
                        default=interval_mean,
                    ),
                ),
            )

        assert hasattr(node, ftor._assigned_property)

    return trie


class AssignOriginTimeNaiveTriePostprocessor(TriePostprocessorBase):
    """Functor to assign origin time property to trie nodes calculated as the
    average of the node's rank and the minimum rank among its children.

    Optionally calculates origin time expected value over this interval for
    a provided prior distribution.
    """

    _assigned_property: str  # property name for assigned origin time
    _prior: PriorBase  # prior expectation for ancestor origin times

    def __init__(
        self: "AssignOriginTimeNaiveTriePostprocessor",
        prior: PriorBase = ArbitraryPrior(),  # ok as kwarg; immutable
        assigned_property: str = "origin_time",
    ) -> None:
        """Initialize functor instance.

        Parameters
        ----------
        prior : PriorBase, default ArbitraryPrior()
            Prior distribution of ancestor origin times.

            Used to calculate interval means.
        assigned_property : str, default "origin_time"
            The property name for the assigned origin time.
        """
        self._assigned_property = assigned_property
        self._prior = prior

    def __call__(
        self: "AssignOriginTimeNaiveTriePostprocessor",
        trie: TrieInnerNode,
        p_differentia_collision: float,
        mutate: bool = False,
        progress_wrap: typing.Callable = lambda x: x,
    ) -> TrieInnerNode:
        """Assign origin times to trie nodes.

        Parameters
        ----------
        trie : TrieInnerNode
            The input trie to be postprocessed.
        p_differentia_collision : float
            Probability of a randomly-generated differentia matching an
            existing differentia.

            Not used in the current implementation.
        mutate : bool, default False
            Are side effects on the input argument `trie` allowed?
        progress_wrap : typing.Callable, optional
            Pass tqdm or equivalent to report progress.

        Returns
        -------
        TrieInnerNode
            The postprocessed trie with assigned origin times.
        """
        if isinstance(trie, TrieInnerNode):
            if not mutate:
                trie = anytree_iterative_deepcopy(
                    trie, progress_wrap=progress_wrap
                )
            return _call_anytree(
                self,
                trie,
                progress_wrap=progress_wrap,
            )
        elif isinstance(trie, (pl.DataFrame, pd.DataFrame)):
            raise NotImplementedError  # pragma: no cover
        else:
            raise TypeError  # pragma: no cover



import typing

import numpy as np

from ._jit import jit


# adapted from https://stackoverflow.com/a/41578614
@jit(nopython=True)
def numpy_index(
    array: np.ndarray, item: typing.Any
) -> typing.Optional[typing.Tuple[int, ...]]:
    """Find the index of the first occurrence of a given item in a Numpy array.

    Parameters
    ----------
    array : numpy.ndarray
        The Numpy array to search.
    item : Any
        The item to search for in the array.

    Returns
    -------
    Tuple[int, ...] or None
        A tuple representing the index of the first occurrence of the given
        item in the array, or None if the item is not found in the array.

    Examples
    --------
    >>> a = np.array([1, 2, 3, 4, 5])
    >>> numpy_index(a, 3)
    (2,)
    >>> numpy_index(a, 6)
    None
    """
    for idx, val in np.ndenumerate(array):
        if val == item:
            return idx
    return None



import shutil
import textwrap

import opytional as opyt

from ._count_leading_blanks import count_leading_blanks


def textwrap_respect_indents(text: str, ncol: int | None = None) -> str:
    """Wrap text, with wrapped lines matching leading indentation.

    Parameters
    ----------
    text : str
        The text to wrap.
    ncol : int, optional
        The maximum width of the wrapped lines.
        If not provided, defaults to the terminal width minus 4, or 60.

    Returns
    -------
    str
        The wrapped text, with wrapped lines matching leading indentation.
    """
    ncol = opyt.or_value(ncol, shutil.get_terminal_size((64, 1)).columns - 4)
    lines = [
        count_leading_blanks(line) * " " + chunk.lstrip().rstrip()
        for line in text.splitlines()
        for chunk in textwrap.wrap(
            line, drop_whitespace=False, replace_whitespace=False, width=ncol
        )
        or [""]  # prevent textwrap from gobbling empty lines
    ]
    return "\n".join(lines)



from hstrat._auxiliary_lib import count_leading_blanks


def test_no_leading_blanks():
    assert count_leading_blanks("hello") == 0


def test_with_leading_blanks():
    assert count_leading_blanks("   hello") == 3


def test_only_blanks():
    assert count_leading_blanks("     ") == 5


def test_empty_string():
    assert count_leading_blanks("") == 0


def test_mixed_whitespace_characters():
    assert count_leading_blanks("\t  hello") == 3  # Assuming tabs count as one


def test_tabs_only():
    assert count_leading_blanks("\t\t\t") == 3  # Assuming tabs count as one


def test_no_leading_but_trailing_whitespace():
    assert count_leading_blanks("hello   ") == 0



import math

from ._UniformPrior import UniformPrior
from ._detail import PriorBase


class ExponentialPrior(PriorBase):
    """Enacts a prior expectation that the generation of the most recent common
    ancestor (MRCA) between extant hereditary stratigraphic columns becomes
    exponentialy less likely with increasing antiquity.

    This prior provides a continuous approximation of the geometric
    distribution of time to MRCA expected under the Wright-Fisher model [1].

    .. [1] Alison Etheridge (2006). Course 11 - Evolution in Fluctuating
    Populations. In Mathematical Statistical Physics (pp. 489-545). Elsevier.
    https://doi.org/10.1016/S0924-8099(06)80048-X

    Notes
    -----
    A static factory function to init an instance with a growth factor
    calculated from contextual information like population size and the number
    of generations elapsed since genesis should be made available in the
    future.

    See Also
    --------
    GeometricPrior
        An exact, discrete analog of this prior.
    """

    _growth_factor: float

    def __init__(self: "ExponentialPrior", growth_factor: float):
        self._growth_factor = growth_factor

    def CalcIntervalProbabilityProxy(
        self: "ExponentialPrior", begin_rank: int, end_rank: int
    ) -> float:
        """Characterize the prior probability of the MRCA generation falling
        within an interval range.

        Parameters
        ----------
        begin_rank : int
            The starting rank of the interval, inclusive.
        end_rank : int
            The ending rank of the interval, exclusive.

        Returns
        -------
        float
            The proxy statistic, proportional to the true estimated interval
            probability of the MRCA value by a fixed (but unspecified) constant
            proportion.
        """
        f = self._growth_factor

        if f == 1.0:
            return UniformPrior().CalcIntervalProbabilityProxy(
                begin_rank, end_rank
            )

        # correction to center interval mean on begin_rank
        # for interval size 1, has numerical precision issues and
        # doesn't appear to correct discretization bias
        # correction = (-f + np.log(f) + 1) / (np.log(f) - f * np.log(f))
        # begin_rank = begin_rank - 1 + correction
        # end_rank = end_rank - 1 + correction

        # simplification: remove 1/log(f) multiplicative constant
        # of integral... constant scaling won't affect weighting result
        return f**end_rank - f**begin_rank

    def CalcIntervalConditionedMean(
        self: "ExponentialPrior", begin_rank: int, end_rank: int
    ) -> float:
        """Calculate the centriod of prior probability mass within an interval
        of possible MRCA generations.

        Parameters
        ----------
        begin_rank : int
            The starting rank of the interval, inclusive.
        end_rank : int
            The ending rank of the interval, exclusive.

        Returns
        -------
        float
            The prior expected generation of MRCA conditioned on the assumption
            that the MRCA falls within the given interval.
        """
        f = self._growth_factor

        if f == 1.0:
            return UniformPrior().CalcIntervalConditionedMean(
                begin_rank, end_rank
            )

        if begin_rank == end_rank - 1:
            return begin_rank
        #
        # \int{f^x \, dx}_a^b = (f^b - f^a) / \log(f)
        #
        # so p(x) = \log(f) * f^x / (f^b - f^a)
        #
        # \int{x f^x \, dx}_a^c
        #    = ( f^a - a f^a \log(f) - f^c + c f^c \log(f) )
        #      / \log^2(f)
        #
        # \int{x p(x) \, dx}_a^b
        #    = (a f**a - b f**b ) / (f**a - f**b)
        #      - (f^b - f^a) / ( (f**a - f**b) \log(f) )
        #    = (a f**a - b f**b ) / (f**a - f**b)
        #      + 1 / \log(f) )
        #

        a = begin_rank
        b = end_rank - 1

        return (a * f**a - b * f**b) / (f**a - f**b) - 1 / math.log(f)

    def SampleIntervalConditionedValue(
        self: "ExponentialPrior", begin_rank: int, end_rank: int
    ) -> int:
        """Sample a generation of the MRCA conditioned on the assumption that
        the MRCA falls within the given interval.

        Parameters
        ----------
        begin_rank : int
            The starting rank of the interval, inclusive.
        end_rank : int
            The ending rank of the interval, exclusive.

        Returns
        -------
        int
            A sampled generation of the MRCA, conditioned on the assumption that
            the MRCA falls within the given interval.
        """
        raise NotImplementedError()



from copy import copy, deepcopy
import itertools as it
import pickle
import random
import tempfile

import pytest

from hstrat import hstrat


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_Clone1(always_store_rank_in_stratum, retention_policy, ordered_store):
    original1 = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    original1_copy1 = deepcopy(original1)
    original1_copy2 = original1.Clone()

    assert original1 == original1_copy1
    assert original1 == original1_copy2
    assert original1_copy1 == original1_copy2

    for first in original1, original1_copy1, original1_copy2:
        for second in original1, original1_copy1, original1_copy2:
            assert hstrat.does_have_any_common_ancestor(first, second)


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_Clone2(always_store_rank_in_stratum, retention_policy, ordered_store):
    original2 = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    original2.DepositStratum()
    original2.DepositStratum()

    original2_copy1 = deepcopy(original2)
    original2_copy2 = original2.Clone()

    for which in original2, original2_copy1, original2_copy2:
        for __ in range(20):
            which.DepositStratum()

    for first in original2, original2_copy1, original2_copy2:
        for second in original2, original2_copy1, original2_copy2:
            assert hstrat.does_have_any_common_ancestor(first, second)


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_Clone3(always_store_rank_in_stratum, ordered_store):
    column = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        initial_stratum_annotation=0,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    population = [column.Clone() for __ in range(3)]

    for _generation in range(100):
        _ = _generation

        for f, s in it.combinations(population, 2):
            assert not f.HasDiscardedStrata()
            assert not s.HasDiscardedStrata()
            assert hstrat.does_have_any_common_ancestor(f, s)
            assert (
                hstrat.get_last_common_stratum_between(
                    f,
                    s,
                )
                is not None
            )

        # advance generation
        population[0] = population[0].Clone()
        for individual in population:
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_Clone4(always_store_rank_in_stratum, ordered_store):
    # regression test for bug with tree store cloning
    column = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        initial_stratum_annotation=0,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    population = [column.Clone() for __ in range(3)]

    for _generation in range(100):
        _ = _generation

        for f, s in it.combinations(population, 2):
            assert not f.HasDiscardedStrata()
            assert hstrat.does_have_any_common_ancestor(f, s)
            assert hstrat.get_last_common_stratum_between(f, s) is not None

        # advance generation
        population[0] = population[0].Clone()
        for individual in population:
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_pickle(always_store_rank_in_stratum, retention_policy, ordered_store):
    original = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    with tempfile.TemporaryDirectory() as tmp_path:
        with open(f"{tmp_path}/data", "wb") as tmp_file:
            pickle.dump(original, tmp_file)

        with open(f"{tmp_path}/data", "rb") as tmp_file:
            reconstituted = pickle.load(tmp_file)
            assert reconstituted == original


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_pickle_with_deposits(retention_policy, ordered_store):
    original = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    for __ in range(100):
        original.DepositStratum()
    with tempfile.TemporaryDirectory() as tmp_path:
        with open(f"{tmp_path}/data", "wb") as tmp_file:
            pickle.dump(original, tmp_file)

        with open(f"{tmp_path}/data", "rb") as tmp_file:
            reconstituted = pickle.load(tmp_file)
            assert reconstituted == original


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_pickle_with_population(
    always_store_rank_in_stratum, retention_policy, ordered_store
):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            always_store_rank_in_stratum=always_store_rank_in_stratum,
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for idx in range(20)
    ]

    for gen in range(10**3):
        population[random.randrange(len(population))] = population[
            random.randrange(len(population))
        ].CloneDescendant()

    for original in population:
        with tempfile.TemporaryDirectory() as tmp_path:
            with open(f"{tmp_path}/data", "wb") as tmp_file:
                pickle.dump(original, tmp_file)

            with open(f"{tmp_path}/data", "rb") as tmp_file:
                reconstituted = pickle.load(tmp_file)
                assert reconstituted == original


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_eq(
    always_store_rank_in_stratum,
    retention_policy,
    ordered_store,
):

    original1 = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    copy1 = deepcopy(original1)
    copy2 = original1.Clone()
    original2 = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )

    assert original1 == copy1
    assert original1 == copy2
    assert copy1 == copy2
    assert original1 != original2
    assert copy1 != original2
    assert copy2 != original2

    copy1.DepositStratum()
    assert original1 != copy1

    original1.DepositStratum()
    assert original1 != copy1


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_annotation(
    always_store_rank_in_stratum, retention_policy, ordered_store
):
    column = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        initial_stratum_annotation=0,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    population = [column.Clone() for __ in range(10)]

    for generation in range(100):
        for f, s in it.combinations(population, 2):
            lb, ub = hstrat.calc_rank_of_mrca_bounds_between(
                f, s, prior="arbitrary"
            )
            assert (
                lb
                <= hstrat.get_last_common_stratum_between(f, s).GetAnnotation()
                < ub
            )
            assert (
                lb
                <= hstrat.get_last_common_stratum_between(s, f).GetAnnotation()
                < ub
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        for individual in population:
            individual.DepositStratum(
                annotation=generation + 1,
            )


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_always_store_rank_in_stratum(retention_policy, ordered_store):
    first = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=True,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    second = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=False,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )

    assert not first._ShouldOmitStratumDepositionRank()
    assert first.GetStratumAtColumnIndex(0).GetDepositionRank() == 0
    if retention_policy.CalcRankAtColumnIndex is not None:
        assert second._ShouldOmitStratumDepositionRank()
        assert second.GetStratumAtColumnIndex(0).GetDepositionRank() is None
    else:
        assert not second._ShouldOmitStratumDepositionRank()
        assert second.GetStratumAtColumnIndex(0).GetDepositionRank() == 0

    for gen in range(100):
        assert hstrat.diff_retained_ranks(first, second) == (set(), set())
        first.DepositStratum()
        second.DepositStratum()


def test_GetNumStrataDeposited():
    column = hstrat.HereditaryStratigraphicColumn()
    for i in range(10):
        assert column.GetNumStrataDeposited() == i + 1
        column.DepositStratum()


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
def test_CloneDescendant(
    always_store_rank_in_stratum,
):
    column = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    assert column.GetNumStrataDeposited() == 1
    descendant = column.CloneDescendant(stratum_annotation="annotation")
    assert descendant.GetNumStrataDeposited() == 2
    assert hstrat.does_have_any_common_ancestor(descendant, column)
    assert column.GetNumStrataDeposited() == 1
    assert (
        descendant.GetStratumAtColumnIndex(
            descendant.GetNumStrataRetained() - 1
        ).GetAnnotation()
        == "annotation"
    )


def test_maximal_retention_policy():
    first = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    second = first.Clone()
    third = first.Clone()

    for gen in range(100):
        assert first.GetNumStrataRetained() == gen + 1
        assert second.GetNumStrataRetained() == gen + 1
        assert third.GetNumStrataRetained() == 1

        assert (
            hstrat.calc_rank_of_mrca_uncertainty_between(
                first, second, prior="arbitrary"
            )
            == 0
        )
        assert (
            hstrat.calc_rank_of_mrca_uncertainty_between(
                first, third, prior="arbitrary"
            )
            == 0
        )

        assert (
            hstrat.calc_ranks_since_mrca_uncertainty_with(
                first, second, prior="arbitrary"
            )
            == 0
        )
        assert (
            hstrat.calc_ranks_since_mrca_uncertainty_with(
                second, first, prior="arbitrary"
            )
            == 0
        )
        assert (
            hstrat.calc_ranks_since_mrca_uncertainty_with(
                first, third, prior="arbitrary"
            )
            == 0
        )
        assert (
            hstrat.calc_ranks_since_mrca_uncertainty_with(
                third, first, prior="arbitrary"
            )
            == 0
        )

        first.DepositStratum()
        second.DepositStratum()
        # no layers deposited onto third


def test_minimal_retention_policy():
    first = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=hstrat.nominal_resolution_algo.Policy(),
    )
    second = first.Clone()
    third = first.Clone()

    for gen in range(100):
        assert first.GetNumStrataRetained() == min(2, gen + 1)
        assert second.GetNumStrataRetained() == min(2, gen + 1)
        assert third.GetNumStrataRetained() == 1

        assert hstrat.calc_rank_of_mrca_uncertainty_between(
            first, second, prior="arbitrary"
        ) == max(0, gen - 1)
        assert (
            hstrat.calc_rank_of_mrca_uncertainty_between(
                first, third, prior="arbitrary"
            )
            == 0
        )

        assert hstrat.calc_ranks_since_mrca_uncertainty_with(
            first,
            second,
            prior="arbitrary",
        ) == max(0, gen - 1)
        assert hstrat.calc_ranks_since_mrca_uncertainty_with(
            second,
            first,
            prior="arbitrary",
        ) == max(0, gen - 1)
        assert (
            hstrat.calc_ranks_since_mrca_uncertainty_with(
                first, third, prior="arbitrary"
            )
            == 0
        )
        assert (
            hstrat.calc_ranks_since_mrca_uncertainty_with(
                third, first, prior="arbitrary"
            )
            == 0
        )

        first.DepositStratum()
        second.DepositStratum()
        # no layers deposited onto third


def test_CalcProbabilityDifferentiaCollision():
    assert (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
        ).CalcProbabilityDifferentiaCollision()
        == 0.5
    )
    assert (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=2,
        ).CalcProbabilityDifferentiaCollision()
        == 0.25
    )
    assert (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=3,
        ).CalcProbabilityDifferentiaCollision()
        == 0.125
    )
    assert (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=8,
        ).CalcProbabilityDifferentiaCollision()
        == 1.0 / 256.0
    )


def test_CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions():
    assert (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
        ).CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1.0 - 0.49
        )
        == 1
    )
    assert (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
        ).CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1.0 - 0.95
        )
        == 5
    )
    assert (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=2,
        ).CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1.0 - 0.95
        )
        == 3
    )
    assert (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=5,
        ).CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1.0 - 0.95
        )
        == 1
    )
    assert (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=64,
        ).CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1.0 - 0.49
        )
        == 1
    )
    assert (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=64,
        ).CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1.0 - 0.95
        )
        == 1
    )
    assert (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=64,
        ).CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1.0 - 0.99
        )
        == 1
    )


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_IterRetainedStrata(
    always_store_rank_in_stratum, retention_policy, ordered_store
):
    column = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    for __ in range(20):
        column.DepositStratum()
        assert [*column.IterRetainedStrata()] == [
            column.GetStratumAtColumnIndex(index)
            for index in range(column.GetNumStrataRetained())
        ]


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_IterRetainedDifferentia(
    always_store_rank_in_stratum, retention_policy, ordered_store
):
    column = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    for __ in range(20):
        column.DepositStratum()
        assert [*column.IterRetainedDifferentia()] == [
            column.GetStratumAtColumnIndex(index).GetDifferentia()
            for index in range(column.GetNumStrataRetained())
        ]


def test_GetColumnIndexOfRank():
    assert (
        hstrat.HereditaryStratigraphicColumn().GetColumnIndexOfRank(1) is None
    )


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_GetStratumAtRank(
    always_store_rank_in_stratum, retention_policy, ordered_store
):
    column = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    for __ in range(20):
        for rank, stratum in zip(
            column.IterRetainedRanks(),
            column.IterRetainedStrata(),
        ):
            assert column.GetStratumAtRank(rank) == stratum
        column.DepositStratum()

    assert hstrat.HereditaryStratigraphicColumn().GetStratumAtRank(1) is None


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_DepositStrata_zero(
    always_store_rank_in_stratum, retention_policy, ordered_store
):
    column = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    for __ in range(20):
        clone = column.Clone()
        column.DepositStrata(0)
        assert clone == column
        column.DepositStratum()


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_DepositStrata_one(
    always_store_rank_in_stratum, retention_policy, ordered_store
):
    c1 = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    c2 = c1.Clone()

    for __ in range(20):
        c1.DepositStrata(1)
        c2.DepositStratum()
        assert all(
            a == b
            for a, b in it.zip_longest(
                c1.IterRetainedRanks(),
                c2.IterRetainedRanks(),
            )
        )
        assert c1.GetNumStrataRetained() == c2.GetNumStrataRetained()
        assert c1.GetNumStrataDeposited() == c2.GetNumStrataDeposited()


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_DepositStrata_several(
    always_store_rank_in_stratum, retention_policy, ordered_store
):
    c1 = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )
    c2 = c1.Clone()

    for __ in range(20):
        step = random.randrange(10)
        c1.DepositStrata(step)
        for __ in range(step):
            c2.DepositStratum()
        assert all(
            a == b
            for a, b in it.zip_longest(
                c1.IterRetainedRanks(),
                c2.IterRetainedRanks(),
            )
        )
        assert c1.GetNumStrataRetained() == c2.GetNumStrataRetained()
        assert c1.GetNumStrataDeposited() == c2.GetNumStrataDeposited()


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_IterRankDifferentiaZip(
    always_store_rank_in_stratum, retention_policy, ordered_store
):
    c1 = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )

    for __ in range(100):
        c1.DepositStratum()
        assert [*c1.IterRankDifferentiaZip()] == [
            *zip(c1.IterRetainedRanks(), c1.IterRetainedDifferentia())
        ]
        iter_ = c1.IterRankDifferentiaZip(copyable=True)
        iter_copy = copy(iter_)
        next(iter_copy)
        assert [*iter_copy] == [
            *zip(c1.IterRetainedRanks(), c1.IterRetainedDifferentia())
        ][1:]
        assert [*iter_] == [
            *zip(c1.IterRetainedRanks(), c1.IterRetainedDifferentia())
        ]


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
def test_CloneNthDescendant_zero(always_store_rank_in_stratum):
    column = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    assert column.GetNumStrataDeposited() == 1
    descendant = column.CloneNthDescendant(0)
    assert column is not descendant
    assert descendant.GetNumStrataDeposited() == 1
    assert hstrat.does_have_any_common_ancestor(descendant, column)
    assert column.GetNumStrataDeposited() == 1


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
def test_CloneNthDescendant_one(always_store_rank_in_stratum):
    column = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    assert column.GetNumStrataDeposited() == 1
    descendant = column.CloneNthDescendant(num_stratum_depositions=1)
    assert column is not descendant
    assert descendant.GetNumStrataDeposited() == 2
    assert hstrat.does_have_any_common_ancestor(descendant, column)
    assert column.GetNumStrataDeposited() == 1


@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
def test_CloneNthDescendant_two(always_store_rank_in_stratum):
    column = hstrat.HereditaryStratigraphicColumn(
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    assert column.GetNumStrataDeposited() == 1
    descendant = column.CloneNthDescendant(num_stratum_depositions=2)
    assert column is not descendant
    assert descendant.GetNumStrataDeposited() == 3
    assert hstrat.does_have_any_common_ancestor(descendant, column)
    assert column.GetNumStrataDeposited() == 1



import random
import typing

import networkx as nx
from networkx.utils import py_random_state as nx_random_state


@nx_random_state("seed")
def random_tree(
    n: int, seed: random.Random = None, create_using: typing.Type = None
) -> nx.Graph:
    """Returns a uniformly random tree on `n` nodes.

    Adapted from NetworkX following deprecation and removal of the function.

    Parameters
    ----------
    n : int
        A positive integer representing the number of nodes in the tree.
    seed : integer, random_state, or None (default)
        Indicator of random number generation state.
        See :ref:`Randomness<randomness>`.
    create_using : NetworkX graph constructor, optional (default=nx.Graph)
        Graph type to create. If graph instance, then cleared before populated.

    Returns
    -------
    NetworkX graph
        A tree, given as an undirected graph, whose nodes are numbers in
        the set {0, …, *n* - 1}.

    Raises
    ------
    NetworkXPointlessConcept
        If `n` is zero (because the null graph is not a tree).

    Notes
    -----
    The current implementation of this function generates a uniformly
    random Prüfer sequence then converts that to a tree via the
    :func:`~networkx.from_prufer_sequence` function. Since there is a
    bijection between Prüfer sequences of length *n* - 2 and trees on
    *n* nodes, the tree is chosen uniformly at random from the set of
    all trees on *n* nodes.

    Examples
    --------
    >>> tree = nx.random_tree(n=10, seed=0)
    >>> nx.write_network_text(tree, sources=[0])
    ╙── 0
        ├── 3
        └── 4
            ├── 6
            │   ├── 1
            │   ├── 2
            │   └── 7
            │       └── 8
            │           └── 5
            └── 9

    >>> tree = nx.random_tree(n=10, seed=0, create_using=nx.DiGraph)
    >>> nx.write_network_text(tree)
    ╙── 0
        ├─╼ 3
        └─╼ 4
            ├─╼ 6
            │   ├─╼ 1
            │   ├─╼ 2
            │   └─╼ 7
            │       └─╼ 8
            │           └─╼ 5
            └─╼ 9
    """
    if n == 0:
        raise nx.NetworkXPointlessConcept("the null graph is not a tree")
    # Cannot create a Prüfer sequence unless `n` is at least two.
    if n == 1:
        utree = nx.empty_graph(1, create_using)
    else:
        sequence = [seed.choice(range(n)) for i in range(n - 2)]
        utree = nx.from_prufer_sequence(sequence)

    if create_using is None:
        tree = utree
    else:
        tree = nx.empty_graph(0, create_using)
        if tree.is_directed():
            # Use a arbitrary root node and dfs to define edge directions
            edges = nx.dfs_edges(utree, source=0)
        else:
            edges = utree.edges

        # Populate the specified graph type
        tree.add_nodes_from(utree.nodes)
        tree.add_edges_from(edges)

    return tree


# NetworkX is distributed with the 3-clause BSD license.

# ::

#    Copyright (C) 2004-2024, NetworkX Developers
#    Aric Hagberg <hagberg@lanl.gov>
#    Dan Schult <dschult@colgate.edu>
#    Pieter Swart <swart@lanl.gov>
#    All rights reserved.

#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are
#    met:

#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.

#      * Neither the name of the NetworkX Developers nor the names of its
#        contributors may be used to endorse or promote products derived
#        from this software without specific prior written permission.

#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#    OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



import typing

import numpy as np

from ..._auxiliary_lib import cmp_approx, pairwise
from ..priors._detail import PriorBase


def estimate_rank_of_mrca_maximum_likelihood(
    coincident_ranks_from_first_disparity_through_first_commonality: typing.Iterator[
        int
    ],
    p_differentia_collision: float,
    prior: PriorBase,
) -> typing.Optional[float]:
    """Estimate the most recent common ancestor (MRCA) at the rank with the
    highest posterior probability.

    Parameters
    ----------
    coincident_ranks_from_first_disparity_through_first_commonality
    : typing.Iterator[int]
        Iterator of integer values indicating the coincident ranks between two
        taxa.

        Inclusive to first disparity and first commonality.
    p_differentia_collision : float
        The multiplicative inverse of the number of possible differentia.
    prior : PriorBase
        Prior expectation for the distribution of MRCA generation
        between hereditary stratigraphic columns.

    Returns
    -------
    typing.Optional[float]
        Estimated rank of the MRCA, or None if the two hereditary stratigraphic
        artifacts definitively share no common ancestor.

    Notes
    -----
    The midpoint rank will be estimated for windows of equal posterior
    probability.
    """

    base = p_differentia_collision

    max_weight, best_expected_rank = -1.0, None
    for (
        num_spurious_collisions,
        (
            end_exclusive,
            begin_inclusive,
        ),
    ) in enumerate(
        pairwise(
            coincident_ranks_from_first_disparity_through_first_commonality
        )
    ):
        cur_weight = (
            base**num_spurious_collisions
            * prior.CalcIntervalProbabilityProxy(
                begin_inclusive, end_exclusive
            )
        )
        assert cur_weight >= 0

        if cur_weight > max_weight:
            max_weight = cur_weight
            # slide to most likely end of interval
            # assumes monotonicity of prior
            best_expected_rank = {
                1: end_exclusive - 1,
                0: np.mean((end_exclusive - 1, begin_inclusive)),
                -1: begin_inclusive,
            }[
                cmp_approx(
                    prior.CalcIntervalConditionedMean(
                        begin_inclusive, end_exclusive
                    ),
                    np.mean((end_exclusive - 1, begin_inclusive)),
                )
            ]

        # if whole rest of the remaining record cannot have higher cur_weight
        # than best, terminate early
        if (
            begin_inclusive
            and base ** (num_spurious_collisions + 1)
            * prior.CalcIntervalProbabilityProxy(
                0,
                begin_inclusive,
            )
            < max_weight
        ):
            break

    assert max_weight > 0.0 or best_expected_rank is None

    # may be None if only one waypoint provided
    return best_expected_rank



import polars as pl


def alifestd_make_ancestor_list_col_polars(
    ids: pl.Series,
    ancestor_ids: pl.Series,
    root_ancestor_token: str = "none",
) -> pl.Series:
    """Translate a column of integer ancestor id values into alife standard
    `ancestor_list` representation.

    The option `root_ancestor_token` will be sandwiched in brackets to create
    the ancestor list entry for genesis organisms. For example, the token
    "None" will yield the entry "[None]" and the token "" will yield the entry
    "[]". Default "none".
    """

    res = ancestor_ids.cast(pl.String)
    res[ids == ancestor_ids] = root_ancestor_token
    return "[" + res + "]"



import math

from .._auxiliary_lib import HereditaryStratigraphicArtifact
from ._calc_probability_differentia_collision_between import (
    calc_probability_differentia_collision_between,
)


def calc_min_implausible_spurious_consecutive_differentia_collisions_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    significance_level: float,
) -> int:
    """Determine amount of evidence required to indicate shared ancestry.

    Calculates how many differentia collisions are required to reject the
    null hypothesis that columns do not share common ancestry at those
    ranks at significance level significance_level.
    """
    assert 0.0 <= significance_level <= 1.0
    assert (
        first.GetStratumDifferentiaBitWidth()
        == second.GetStratumDifferentiaBitWidth()
    )

    if significance_level * first.GetStratumDifferentiaBitWidth() > 1:
        return 1

    log_base = calc_probability_differentia_collision_between(first, second)
    if significance_level >= log_base:
        return 1
    else:
        return int(math.ceil(math.log(significance_level, log_base)))



from collections import Counter

import pandas as pd

from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_count_unifurcations(phylogeny_df: pd.DataFrame) -> int:
    """Count how many inner nodes have exactly one descendant node.

    Only supports asexual phylogenies.
    """
    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df)
    if "ancestor_id" not in phylogeny_df.columns:
        raise ValueError(
            "alifestd_count_unifurcations only supports asexual phylogenies.",
        )
    except_roots = phylogeny_df["ancestor_id"] != phylogeny_df["id"]
    ancestor_counts = Counter(phylogeny_df.loc[except_roots, "ancestor_id"])
    return sum(v == 1 for v in ancestor_counts.values())



import pandas as pd

from ..genome_instrumentation import HereditaryStratigraphicColumn


def col_to_dataframe(column: HereditaryStratigraphicColumn) -> pd.DataFrame:
    """Create a pandas `DataFrame` with retained strata as rows."""
    records = [
        {
            "index": index,
            "rank": column.GetRankAtColumnIndex(index),
            "differentia": stratum.GetDifferentia(),
        }
        for index, stratum in enumerate(column.IterRetainedStrata())
    ]
    return pd.DataFrame.from_records(records).astype(int)



import pytest

from hstrat import genome_instrumentation, hstrat
from hstrat._auxiliary_lib import flat_len


@pytest.mark.parametrize(
    "impl",
    [genome_instrumentation.HereditaryStratigraphicColumn],
    # TODO
    # genome_instrumentation._HereditaryStratigraphicColumn_.impls,
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
        None,
    ],
)
@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 64],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 16, 32, 64],
)
def test_col_to_int(
    impl,
    retention_policy,
    ordered_store,
    always_store_rank_in_stratum,
    num_deposits,
    differentia_bit_width,
    caplog,
):
    column = impl(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_bit_width,
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    for __ in range(num_deposits):
        column.DepositStratum()

    packet = hstrat.col_to_int(column)
    assert packet == hstrat.col_to_int(column)
    assert packet != hstrat.col_to_int(
        column, num_strata_deposited_byte_order="little"
    )
    assert packet != hstrat.col_to_int(
        column, num_strata_deposited_byte_width=42
    )
    assert isinstance(packet, int)
    assert packet.bit_length() <= 8 * (
        (column.GetNumStrataRetained() * differentia_bit_width + 7 + 1) // 8
        + 1  # possible num padding bits header byte
        + 4
    )


def test_col_from_int():
    value = 0x100000008F6
    column = hstrat.col_from_int(
        value,
        differentia_bit_width=1,
        num_strata_deposited_byte_order="big",
        num_strata_deposited_byte_width=4,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    assert not column._always_store_rank_in_stratum
    assert column.GetNumStrataDeposited() == 8
    assert flat_len(column.IterRetainedDifferentia()) == 8
    assert flat_len(column.IterRetainedRanks()) == 8

    assert [*column.IterRankDifferentiaZip()] == [
        (0, 1),
        (1, 1),
        (2, 1),
        (3, 1),
        (4, 0),
        (5, 1),
        (6, 1),
        (7, 0),
    ]


def test_col_from_int_differentiae_byte_bit_order():
    value = 0x100000008F7
    column = hstrat.col_from_int(
        value,
        differentia_bit_width=1,
        differentiae_byte_bit_order="little",
        num_strata_deposited_byte_order="big",
        num_strata_deposited_byte_width=4,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    assert not column._always_store_rank_in_stratum
    assert column.GetNumStrataDeposited() == 8
    assert flat_len(column.IterRetainedDifferentia()) == 8
    assert flat_len(column.IterRetainedRanks()) == 8

    assert [*column.IterRankDifferentiaZip()] == [
        (0, 1),
        (1, 1),
        (2, 1),
        (3, 0),
        (4, 1),
        (5, 1),
        (6, 1),
        (7, 1),
    ]


def test_col_from_int_endian():
    value = 0x108000000F7
    column = hstrat.col_from_int(
        value,
        differentia_bit_width=1,
        num_strata_deposited_byte_order="little",
        num_strata_deposited_byte_width=4,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    assert not column._always_store_rank_in_stratum
    assert column.GetNumStrataDeposited() == 8
    assert flat_len(column.IterRetainedDifferentia()) == 8
    assert flat_len(column.IterRetainedRanks()) == 8

    assert [*column.IterRankDifferentiaZip()] == [
        (0, 1),
        (1, 1),
        (2, 1),
        (3, 1),
        (4, 0),
        (5, 1),
        (6, 1),
        (7, 1),
    ]


def test_col_from_int_byte_width():
    value = 0x8F6
    column = hstrat.col_from_int(
        value,
        differentia_bit_width=1,
        num_strata_deposited_byte_width=4,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
        value_byte_width=5,
    )
    assert not column._always_store_rank_in_stratum
    assert column.GetNumStrataDeposited() == 8
    assert flat_len(column.IterRetainedDifferentia()) == 8
    assert flat_len(column.IterRetainedRanks()) == 8

    assert [*column.IterRankDifferentiaZip()] == [
        (0, 1),
        (1, 1),
        (2, 1),
        (3, 1),
        (4, 0),
        (5, 1),
        (6, 1),
        (7, 0),
    ]


@pytest.mark.parametrize(
    "impl",
    [genome_instrumentation.HereditaryStratigraphicColumn],
    # TODO
    # genome_instrumentation._HereditaryStratigraphicColumn_.impls,
)
@pytest.mark.parametrize("byte_order", ["big", "little"])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
        None,
    ],
)
@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 64],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 16, 32, 64],
)
def test_col_to_int_then_from_int(
    impl,
    byte_order,
    retention_policy,
    ordered_store,
    always_store_rank_in_stratum,
    num_deposits,
    differentia_bit_width,
    caplog,
):
    column = impl(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_bit_width,
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    for __ in range(num_deposits):
        column.DepositStratum()

    assert hstrat.col_to_int(column) == hstrat.col_to_int(column)
    reconstituted = hstrat.col_from_int(
        hstrat.col_to_int(column, num_strata_deposited_byte_order=byte_order),
        differentia_bit_width=differentia_bit_width,
        num_strata_deposited_byte_order=byte_order,
        stratum_retention_policy=retention_policy,
    )
    assert reconstituted._ShouldOmitStratumDepositionRank()
    if (
        ordered_store == hstrat.HereditaryStratumOrderedStoreList
        and impl == genome_instrumentation.HereditaryStratigraphicColumn
        and not always_store_rank_in_stratum
    ):
        assert reconstituted == column
    else:
        target_records = hstrat.col_to_records(column)
        # always_store_rank_in_stratum setting is lost
        if "stratum_deposition_ranks" in target_records:
            del target_records["stratum_deposition_ranks"]
        assert hstrat.col_to_records(reconstituted) == target_records



import anytree
import pytest

from hstrat._auxiliary_lib import anytree_cardinality, is_in
import hstrat.phylogenetic_inference.tree._impl as impl


@pytest.fixture
def subtrie():
    # (rank None, diff None) @ B6A
    # └── (rank 0, diff 2) @ BtQ
    #     ├── (rank 3, diff 31) @ jg
    #     └── (rank 3, diff 1) @ BUQ
    #         ├── spice @ 0A
    #         ├── (rank 4, diff 17) @ BxA
    #         │   └── (rank 5, diff 11) @ Gg
    #         │       ├── (rank 6, diff 19) @ Biw
    #         │       └── (rank 6, diff 21) @ HA
    #         │           └── baa @ B-Q
    #         ├── (rank 4, diff 16) @ B7Q
    #         │   └── (rank 5, diff 11) @ sA
    #         │       └── foo @ BtQ
    #         └── (rank 4, diff 17) @ Brw
    #             ├── (rank 6, diff 19) @ BLg
    #             │   └── baz @ B0g
    #             └── (rank 6, diff 19) @ BoA
    #                 └── bar @ B8A

    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    genesis_node = impl.TrieInnerNode(rank=0, differentia=2, parent=root)
    _sibling_node = impl.TrieInnerNode(
        rank=3, differentia=31, parent=genesis_node
    )
    _ = _sibling_node
    focal_node = impl.TrieInnerNode(rank=3, differentia=1, parent=genesis_node)
    impl.TrieLeafNode(focal_node, "spice")
    child1 = impl.TrieInnerNode(rank=4, differentia=17, parent=focal_node)
    child2 = impl.TrieInnerNode(rank=4, differentia=16, parent=focal_node)
    child3 = impl.TrieInnerNode(rank=4, differentia=17, parent=focal_node)
    grandchild1 = impl.TrieInnerNode(rank=5, differentia=11, parent=child1)
    grandchild2 = impl.TrieInnerNode(rank=5, differentia=11, parent=child2)
    impl.TrieLeafNode(grandchild2, "foo")
    greatgrandchild1 = impl.TrieInnerNode(
        rank=6, differentia=19, parent=grandchild1
    )
    greatgrandchild1 = impl.TrieInnerNode(
        rank=6, differentia=21, parent=grandchild1
    )
    impl.TrieLeafNode(greatgrandchild1, "baa")
    greatgrandchild2 = impl.TrieInnerNode(
        rank=6, differentia=19, parent=child3
    )
    impl.TrieLeafNode(greatgrandchild2, "baz")
    greatgrandchild3 = impl.TrieInnerNode(
        rank=6, differentia=19, parent=child3
    )
    impl.TrieLeafNode(greatgrandchild3, "bar")

    return focal_node


def test_is_genesis_of_allele():
    root_node = impl.TrieInnerNode(rank=None, differentia=None)
    inner_node = impl.TrieInnerNode(rank=1, differentia=2, parent=root_node)
    assert inner_node.IsAnOriginationOfAllele(1, 2)
    assert not inner_node.IsAnOriginationOfAllele(1, 3)
    assert not inner_node.IsAnOriginationOfAllele(2, 2)

    assert not root_node.IsAnOriginationOfAllele(1, 2)


def test_find_geneses_of_allele_self(subtrie):
    geneses = [*subtrie.FindOriginationsOfAllele(3, 1)]
    assert len(geneses) == 1
    assert geneses[0] is subtrie


def test_find_geneses_of_allele_singleton(subtrie):
    geneses = [*subtrie.FindOriginationsOfAllele(6, 21)]
    assert len(geneses) == 1
    assert geneses[0].IsAnOriginationOfAllele(6, 21)


def test_find_geneses_of_allele_multiple(subtrie):
    geneses = [*subtrie.FindOriginationsOfAllele(6, 19)]
    assert len(geneses) == 3

    for genesis in geneses:
        assert genesis.IsAnOriginationOfAllele(6, 19)

    geneses = [*subtrie.FindOriginationsOfAllele(4, 17)]
    assert len(geneses) == 2

    for genesis in geneses:
        assert genesis.IsAnOriginationOfAllele(4, 17)

    geneses = [*subtrie.FindOriginationsOfAllele(5, 11)]
    assert len(geneses) == 2

    for genesis in geneses:
        assert genesis.IsAnOriginationOfAllele(5, 11)


def test_find_geneses_of_allele_failure(subtrie):
    geneses = [*subtrie.FindOriginationsOfAllele(3, 19)]
    assert len(geneses) == 0

    geneses = [*subtrie.FindOriginationsOfAllele(3, 31)]
    assert len(geneses) == 0

    geneses = [*subtrie.FindOriginationsOfAllele(11, 31)]
    assert len(geneses) == 0

    geneses = [*subtrie.FindOriginationsOfAllele(5, 21)]
    assert len(geneses) == 0


def test_get_deepest_consecutive_shared_allele_genesis_with_empty_iterator(
    subtrie,
):
    (
        genesis,
        leftover_iter,
    ) = subtrie.GetDeepestCongruousAlleleOrigination(iter([]))
    assert genesis is subtrie
    assert [*leftover_iter] == []


def test_get_deepest_consecutive_shared_allele_genesis_singleton(
    subtrie,
):
    # test with a non-empty iterator
    alleles = [(4, 16), (5, 11)]
    (
        genesis,
        leftover_iter,
    ) = subtrie.GetDeepestCongruousAlleleOrigination(iter(alleles))
    assert genesis.IsAnOriginationOfAllele(5, 11)
    assert genesis.parent.IsAnOriginationOfAllele(4, 16)
    assert [*leftover_iter] == []

    alleles = [(4, 16)]
    (
        genesis,
        leftover_iter,
    ) = subtrie.GetDeepestCongruousAlleleOrigination(iter(alleles))
    assert genesis.IsAnOriginationOfAllele(4, 16)
    assert [*leftover_iter] == []

    alleles = [(4, 16), (5, 11), (7, 28), (8, 19)]
    (
        genesis,
        leftover_iter,
    ) = subtrie.GetDeepestCongruousAlleleOrigination(iter(alleles))
    assert genesis.IsAnOriginationOfAllele(5, 11)
    assert genesis.parent.IsAnOriginationOfAllele(4, 16)
    assert [*leftover_iter] == alleles[2:]

    alleles = [(4, 16), (6, 19)]
    (
        genesis,
        leftover_iter,
    ) = subtrie.GetDeepestCongruousAlleleOrigination(iter(alleles))
    assert genesis.IsAnOriginationOfAllele(4, 16)
    assert [*leftover_iter] == alleles[1:]


def test_get_deepest_consecutive_shared_allele_genesis_tiebreaker(
    subtrie,
):
    # test with a non-empty iterator
    alleles = [(4, 17), (6, 19), (7, 22)]
    (
        genesis,
        leftover_iter,
    ) = subtrie.GetDeepestCongruousAlleleOrigination(iter(alleles))
    assert genesis.IsAnOriginationOfAllele(6, 19)
    assert [*leftover_iter] == alleles[2:]

    (
        genesis2,
        leftover_iter,
    ) = subtrie.GetDeepestCongruousAlleleOrigination(iter(alleles))
    assert genesis2 is genesis
    assert [*leftover_iter] == alleles[2:]

    alleles = [(4, 17)]
    (
        genesis,
        leftover_iter,
    ) = subtrie.GetDeepestCongruousAlleleOrigination(iter(alleles))
    assert is_in(genesis2, anytree.PostOrderIter(genesis))
    assert [*leftover_iter] == []


def test_insert_taxon_canopy1(subtrie):
    previous_node_count = anytree_cardinality(subtrie)
    alleles = [(4, 17)]
    (
        genesis,
        leftover_iter,
    ) = subtrie.GetDeepestCongruousAlleleOrigination(iter(alleles))
    assert [*leftover_iter] == []

    inserted = genesis.InsertTaxon("fab", iter([]))
    assert inserted.taxon == "fab"
    assert inserted.parent is genesis
    assert inserted.parent.IsAnOriginationOfAllele(4, 17)
    assert anytree_cardinality(subtrie) == previous_node_count + 1
    previous_node_count = anytree_cardinality(subtrie)

    inserted = genesis.InsertTaxon("far", iter([(9, 12)]))
    assert inserted.taxon == "far"
    assert inserted.parent.parent is genesis
    assert inserted.parent.IsAnOriginationOfAllele(9, 12)
    assert anytree_cardinality(subtrie) == previous_node_count + 2
    previous_node_count = anytree_cardinality(subtrie)


def test_insert_taxon_canopy2(subtrie):
    previous_node_count = anytree_cardinality(subtrie)
    alleles = [(4, 17), (6, 19)]
    (
        genesis,
        leftover_iter,
    ) = subtrie.GetDeepestCongruousAlleleOrigination(iter(alleles))
    assert [*leftover_iter] == []
    inserted = genesis.InsertTaxon("fab", iter([]))
    assert inserted.taxon == "fab"
    assert inserted.parent is genesis
    assert anytree_cardinality(subtrie) == previous_node_count + 1
    previous_node_count = anytree_cardinality(subtrie)

    inserted = genesis.InsertTaxon("far", iter([(7, 12)]))
    assert inserted.taxon == "far"
    assert inserted.parent.parent is genesis
    assert anytree_cardinality(subtrie) == previous_node_count + 2


def test_insert_taxon_canopy3(subtrie):
    previous_node_count = anytree_cardinality(subtrie)
    alleles = [(4, 17), (6, 19)]
    (
        genesis,
        leftover_iter,
    ) = subtrie.GetDeepestCongruousAlleleOrigination(iter(alleles))
    assert [*leftover_iter] == []

    inserted = genesis.InsertTaxon("gg", iter([(7, 12), (8, 20)]))
    assert inserted.taxon == "gg"
    assert inserted.parent.parent.parent is genesis
    assert inserted.parent.IsAnOriginationOfAllele(8, 20)
    assert inserted.parent.parent.IsAnOriginationOfAllele(7, 12)
    assert anytree_cardinality(subtrie) == previous_node_count + 3


def test_insert_taxon_internal1(subtrie):
    previous_node_count = anytree_cardinality(subtrie)
    alleles = [(4, 17)]
    (
        genesis,
        leftover_iter,
    ) = subtrie.GetDeepestCongruousAlleleOrigination(iter(alleles))
    assert [*leftover_iter] == []
    children_before = [*genesis.children]

    inserted = genesis.InsertTaxon("fab", iter([(5, 11)]))
    assert inserted.taxon_label == "fab"
    assert inserted.parent.IsAnOriginationOfAllele(5, 11)
    assert is_in(inserted.parent, children_before)
    assert inserted.parent.parent is genesis
    assert anytree_cardinality(subtrie) == previous_node_count + 1


def test_insert_taxon_internal2(subtrie):
    previous_node_count = anytree_cardinality(subtrie)
    alleles = [(4, 17)]
    (
        genesis,
        leftover_iter,
    ) = subtrie.GetDeepestCongruousAlleleOrigination(iter(alleles))
    assert [*leftover_iter] == []
    descendants_before = [*anytree.PostOrderIter(genesis)]

    inserted = genesis.InsertTaxon("fab", iter([(5, 11), (6, 21)]))
    assert inserted.taxon_label == "fab"
    assert inserted.parent.IsAnOriginationOfAllele(6, 21)
    assert anytree_cardinality(subtrie) == previous_node_count + 1

    previous_node_count = anytree_cardinality(subtrie)
    descendants_before = [*anytree.PostOrderIter(genesis)]
    inserted = genesis.InsertTaxon("fan", iter([(5, 11), (6, 21)]))
    assert inserted.taxon_label == "fan"
    assert is_in(inserted.parent, descendants_before)
    assert anytree_cardinality(subtrie) == previous_node_count + 1


def test_insert_taxon_internal3(subtrie):
    alleles = [(4, 17)]
    (
        genesis,
        leftover_iter,
    ) = subtrie.GetDeepestCongruousAlleleOrigination(iter(alleles))
    assert [*leftover_iter] == []
    inserted = genesis.InsertTaxon("fab", iter([(5, 11), (6, 21)]))
    assert inserted.taxon_label == "fab"
    previous_node_count = anytree_cardinality(subtrie)
    descendants_before = [*anytree.PostOrderIter(genesis)]

    inserted = genesis.InsertTaxon(
        "bat", iter([(5, 11), (6, 21), (10, 10), (11, 19)])
    )
    assert inserted.taxon_label == "bat"
    assert inserted.parent.IsAnOriginationOfAllele(11, 19)
    assert inserted.parent.parent.IsAnOriginationOfAllele(10, 10)
    assert inserted.parent.parent.parent.IsAnOriginationOfAllele(6, 21)
    assert not is_in(inserted.parent, descendants_before)
    assert not is_in(inserted.parent.parent, descendants_before)
    assert is_in(inserted.parent.parent.parent, descendants_before)
    assert anytree_cardinality(subtrie) == previous_node_count + 3


def test_taxon_and_taxon_label(subtrie):
    taxon_labels = [
        node.taxon_label for node in anytree.PostOrderIter(subtrie)
    ]
    assert len(taxon_labels) == len(set(taxon_labels))

    for node in anytree.PostOrderIter(subtrie):
        assert node.taxon == node.taxon_label
        assert isinstance(node.taxon, str)


def test_eq_empty():
    n1 = impl.TrieInnerNode()
    m1 = impl.TrieInnerNode()
    assert n1 == m1


def test_eq_equal():
    n1 = impl.TrieInnerNode()
    n2 = impl.TrieInnerNode(parent=n1, rank=12, differentia=50)
    n3 = impl.TrieInnerNode(parent=n1, rank=2, differentia=36)
    n4 = impl.TrieInnerNode(parent=n2, rank=30, differentia=145)
    n5 = impl.TrieInnerNode(parent=n2, rank=6, differentia=101)
    n6 = impl.TrieInnerNode(parent=n5, rank=7, differentia=43)
    impl.TrieLeafNode(parent=n6, taxon_label="cat")
    impl.TrieLeafNode(parent=n3, taxon_label="dog")
    impl.TrieLeafNode(parent=n4, taxon_label="elephant")
    impl.TrieLeafNode(parent=n4, taxon_label="rhino")

    m1 = impl.TrieInnerNode()
    m3 = impl.TrieInnerNode(parent=m1, rank=2, differentia=36)
    m2 = impl.TrieInnerNode(parent=m1, rank=12, differentia=50)
    m5 = impl.TrieInnerNode(parent=m2, rank=6, differentia=101)
    m6 = impl.TrieInnerNode(parent=m5, rank=7, differentia=43)
    m4 = impl.TrieInnerNode(parent=m2, rank=30, differentia=145)
    impl.TrieLeafNode(parent=m3, taxon_label="dog")
    impl.TrieLeafNode(parent=m4, taxon_label="rhino")
    impl.TrieLeafNode(parent=m4, taxon_label="elephant")
    impl.TrieLeafNode(parent=m6, taxon_label="cat")

    assert m1 == n1


def test_not_eq_lacking_leaf():
    n1 = impl.TrieInnerNode()
    n2 = impl.TrieInnerNode(parent=n1, rank=12, differentia=50)
    n3 = impl.TrieInnerNode(parent=n1, rank=2, differentia=36)
    impl.TrieLeafNode(parent=n2, taxon_label="rhino")
    impl.TrieLeafNode(parent=n3, taxon_label="apple")

    m1 = impl.TrieInnerNode()
    m2 = impl.TrieInnerNode(parent=m1, rank=12, differentia=50)
    impl.TrieInnerNode(parent=m1, rank=2, differentia=36)
    impl.TrieLeafNode(parent=m2, taxon_label="rhino")
    assert n1 != m1


def test_not_eq_lacking_inner():
    n1 = impl.TrieInnerNode()
    n2 = impl.TrieInnerNode(parent=n1, rank=12, differentia=50)
    impl.TrieLeafNode(parent=n2, taxon_label="rhino")

    m1 = impl.TrieInnerNode()
    m2 = impl.TrieInnerNode(parent=m1, rank=12, differentia=50)
    impl.TrieInnerNode(parent=m1, rank=2, differentia=36)
    impl.TrieLeafNode(parent=m2, taxon_label="rhino")
    assert n1 != m1


def test_not_eq_different_data():
    n1 = impl.TrieInnerNode()
    n2 = impl.TrieInnerNode(parent=n1, rank=12, differentia=50)
    n3 = impl.TrieInnerNode(parent=n1, rank=1, differentia=35)
    impl.TrieLeafNode(parent=n2, taxon_label="rhino")
    impl.TrieLeafNode(parent=n3, taxon_label="apple")

    m1 = impl.TrieInnerNode()
    m2 = impl.TrieInnerNode(parent=m1, rank=12, differentia=50)
    m3 = impl.TrieInnerNode(parent=m1, rank=2, differentia=36)
    impl.TrieLeafNode(parent=m2, taxon_label="rhino")
    impl.TrieLeafNode(parent=m3, taxon_label="apple")
    assert n1 != m1


def test_inner_children(subtrie):
    assert len([*subtrie.inner_children]) == 3
    assert len(set(map(id, subtrie.inner_children))) == 3
    assert all(
        isinstance(node, impl.TrieInnerNode) for node in subtrie.inner_children
    )
    assert all(node.parent is subtrie for node in subtrie.inner_children)


def test_outer_children(subtrie):
    assert len([*subtrie.outer_children]) == 1
    assert len(set(map(id, subtrie.outer_children))) == 1
    assert all(
        isinstance(node, impl.TrieLeafNode) for node in subtrie.outer_children
    )
    assert all(node.parent is subtrie for node in subtrie.outer_children)



import pandas as pd

from hstrat._auxiliary_lib import (
    alifestd_is_chronologically_sorted,
    alifestd_make_empty,
)


def test_is_chronologically_sorted_empty():
    df = alifestd_make_empty()
    df["origin_time"] = []
    assert alifestd_is_chronologically_sorted(df)


def test_is_chronologically_sorted_missing_col():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
        }
    )
    assert not alifestd_is_chronologically_sorted(phylogeny_df)


def test_is_chronologically_sorted():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "origin_time": [0, 1, 2, 2, 5, 6],
        }
    )
    assert alifestd_is_chronologically_sorted(phylogeny_df)


def test_is_not_chronologically_sorted():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "origin_time": [2, 1, 2, 2, 5, 6],
        }
    )
    assert not alifestd_is_chronologically_sorted(phylogeny_df)


def test_is_chronologically_sorted_destruction_time():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "destruction_time": [0, 1, 2, 2, 5, 6],
        }
    )
    assert alifestd_is_chronologically_sorted(
        phylogeny_df, how="destruction_time"
    )


def test_is_not_chronologically_sorted_destruction_time():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "destruction_time": [2, 1, 2, 2, 5, 6],
        }
    )
    assert not alifestd_is_chronologically_sorted(
        phylogeny_df, how="destruction_time"
    )



import pandas as pd

from ._alifestd_is_sexual import alifestd_is_sexual


def alifestd_is_asexual(phylogeny_df: pd.DataFrame) -> bool:
    """Do all organisms in the phylogeny have one or no immediate ancestor?

    Input dataframe is not mutated by this operation.
    """
    return "ancestor_id" in phylogeny_df or not alifestd_is_sexual(
        phylogeny_df
    )
